from unittest import TestCase
from cilantro_ee.utils import int_to_bytes, bytes_to_int
import secrets
import random


MAX_INT = pow(2, 64)


class TestIntByteUtils(TestCase):

    def test_int_to_bytes(self):
        b = int_to_bytes(12351)
        self.assertTrue(type(b) is bytes)

    def test_bytes_to_int(self):
        for byte_size in range(8 + 1):
            i = bytes_to_int(secrets.token_bytes(byte_size))
            self.assertTrue(type(i) is int)

    def test_inverse_operations(self):
        # We loop hella times cuz ya boy is paranoid it may not work for odds/integers in different ranges
        for _ in range(100):
            i = random.randint(0, MAX_INT)
            clone = bytes_to_int(int_to_bytes(i))
            self.assertEqual(i, clone)




from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('1-0-0.json')

from cilantro_ee.utils.factory import NodeFactory
from cilantro_ee.constants.testnet import TESTNET_MASTERNODES
from cilantro_ee.services.storage.vkbook import VKBook
import os

# overwrite_logger_level(logging.WARNING)
# overwrite_logger_level(21)
# overwrite_logger_level(11)

os.environ["NONCE_DISABLED"] = "1"
os.environ["MN_MOCK"] = "1"


print("VKBook mns {}".format(VKBook.get_masternodes()))
print("TESTNET MASTERNODES {}".format(TESTNET_MASTERNODES))

ip = '127.0.0.1'
sk = TESTNET_MASTERNODES[0]['sk']
NodeFactory.run_masternode(ip=ip, signing_key=sk, reset_db=True)



from cilantro_ee.constants.zmq_filters import TRANSACTION_FILTER
from cilantro_ee.constants.ports import MN_TX_PUB_PORT
from cilantro_ee.constants.batcher import BATCHER_SLEEP_INTERVAL
from cilantro_ee.constants.batcher import MAX_TXN_SUBMISSION_DELAY
from cilantro_ee.constants.batcher import MAX_TXNS_PER_SUB_BLOCK

from cilantro_ee.core.utils.worker import Worker
from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.crypto.wallet import Wallet, _verify
from cilantro_ee.core.utils.transaction import transaction_is_valid, TransactionException
from cilantro_ee.core.nonces import NonceManager
from cilantro_ee.core.sockets.services import AsyncInbox, SocketStruct, Protocols
from multiprocessing import Queue
import zmq.asyncio
import asyncio
import time
import hashlib

IPC_ID = '/tmp/masternode-input-hash-inbox'
IPC_PORT = 6967

'''
Dynamically rate limits transaction batches so that the Trnasaction Batcher can just await responses and not have to
worry about DDOS attacked or TX flooding.
'''


class NewTransactionBatcher:
    def __init__(self, publisher_ip: str, wallet: Wallet, ctx: zmq.asyncio.Context, queue: Queue=Queue(),ipc_id: str=IPC_ID):
        self.wallet = wallet
        self.queue = queue
        self.ctx = ctx
        self.publisher_ip = publisher_ip
        self.ipc_id = SocketStruct(protocol=Protocols.ICP, id=ipc_id)

        self.ready = False
        self.running = False

        self.input_hash_inbox = InputHashInbox(parent=self, socket_id=self.ipc_id, wallet=wallet, ctx=self.ctx)

        self.batcher = RateLimitingBatcher(self.queue, self.wallet,
                                           BATCHER_SLEEP_INTERVAL,
                                           MAX_TXNS_PER_SUB_BLOCK,
                                           MAX_TXN_SUBMISSION_DELAY)

        self.publisher = self.ctx.socket(zmq.PUB)
        self.publisher.bind(f'tcp://{self.publisher_ip}:{MN_TX_PUB_PORT}')

    async def start(self):
        self.running = True
        asyncio.ensure_future(self.input_hash_inbox.serve())

        while not self.ready and self.running:
            await asyncio.sleep(0)

        asyncio.ensure_future(self.compose_transactions())

    async def compose_transactions(self):
        sub_filter = TRANSACTION_FILTER.encode()

        while self.running:
            mtype, msg = await self.batcher.get_next_batch_packed()
            self.publisher.send(mtype + msg)

    def stop(self):
        self.running = False
        self.input_hash_inbox.stop()
        self.batcher.running = False


class InputHashInbox(AsyncInbox):
    def __init__(self, parent: NewTransactionBatcher, *args, **kwargs):
        self.parent = parent
        super().__init__(*args, **kwargs)

    async def handle_msg(self, _id, msg):
        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(message=msg)

        if not is_verified:
            return

        if msg_type == MessageType.BURN_INPUT_HASHES:
            self.parent.batcher.remove_batch_ids(msg.inputHashes)

        elif msg_type == MessageType.READY:
            self.parent.ready = True


class RateLimitingBatcher:
    def __init__(self, queue, wallet, sleep_interval,
                 max_batch_size, max_txn_delay):

        self.queue = queue
        self.wallet = wallet
        self.batcher_sleep_interval = sleep_interval
        self.max_batch_size = max_batch_size
        self.max_txn_submission_delay = max_txn_delay

        self.driver = NonceManager()
        self.num_batches_sent = 0
        self.txn_delay = 0
        self.tasks = []
        self.sent_batch_ids = []

        self.running = True

    def add_batch_id(self, batch_id):
        self.sent_batch_ids.append(batch_id)
        self.num_batches_sent += 1

    # async def remove_batch_ids(self, batch_ids):
    def remove_batch_ids(self, batch_ids):
        self.num_batches_sent -= 1

        for id in batch_ids:
            while id in self.sent_batch_ids:
                self.sent_batch_ids.pop(0)

        list_len = len(self.sent_batch_ids)

        if list_len < self.num_batches_sent:
            self.num_batches_sent = list_len

    def ready_for_next_batch(self):
        num_txns = self.queue.qsize()

        if num_txns > 0:
            self.txn_delay += 1

        if self.num_batches_sent >= 2 or \
                (self.num_batches_sent > 0 and
                 (num_txns < self.max_batch_size or self.txn_delay < self.max_txn_submission_delay)):
            return False

        return True

    def get_next_batch_size(self):
        num_txns = self.queue.qsize()

        if (num_txns >= self.max_batch_size) or \
           (self.txn_delay >= self.max_txn_submission_delay):

            self.txn_delay = 0
            return min(num_txns, self.max_batch_size)

        return 0

    def get_txn_list(self, batch_size):
        tx_list = []
        for _ in range(batch_size):
            # Get a transaction from the queue
            tx = self.queue.get()


            # Make sure that the transaction is valid
            # this is better done at webserver level before packing and putting it into the queue - raghu todo
            try:
                transaction_is_valid(tx=tx[1],
                                     expected_processor=self.wallet.verifying_key(),
                                     driver=self.driver,
                                     strict=True)
            except TransactionException as e:
                continue

            tx_list.append(tx[1])

        return tx_list

    def pack_txn_list(self, tx_list):
        h = hashlib.sha3_256()
        for tx in tx_list:
            # Hash it
            tx_bytes = tx.as_builder().to_bytes_packed()
            h.update(tx_bytes)

        # Add a timestamp
        timestamp = time.time()
        h.update('{}'.format(timestamp).encode())
        inputHash = h.digest()

        # Make explicit, only being used once
        self.add_batch_id(inputHash)

        # Sign the message for verification
        signature = self.wallet.sign(inputHash)

        mtype, msg = Message.get_message_packed(
                         MessageType.TRANSACTION_BATCH,
                         transactions=[t for t in tx_list], timestamp=timestamp,
                         signature=signature, inputHash=inputHash,
                         sender=self.wallet.verifying_key())

        # self.log.debug1("Send {} transactions with hash {} and timestamp {}"
                        # .format(len(tx_list), inputHash.hex(), timestamp))

        return mtype, msg

    async def get_next_batch_packed(self):
        while not self.ready_for_next_batch() and self.running:
            await asyncio.sleep(self.batcher_sleep_interval)

            # Add to interval here... easier

        # Make this explicit. not being used elsewhere
        num_txns = self.get_next_batch_size()

        tx_list = self.get_txn_list(num_txns)

        mtype, msg = self.pack_txn_list(tx_list)

        return mtype, msg


class TransactionBatcher(Worker):
    def __init__(self, ip, signing_key, queue=Queue(), ipc_ip=IPC_ID, ipc_port=IPC_PORT, *args, **kwargs):
        super().__init__(signing_key=signing_key, *args, **kwargs)
        self.ip = ip
        self.ipc_ip = ipc_ip
        self.ipc_port = ipc_port

        self.batcher = RateLimitingBatcher(queue, self.wallet,
                                           BATCHER_SLEEP_INTERVAL,
                                           MAX_TXNS_PER_SUB_BLOCK,
                                           MAX_TXN_SUBMISSION_DELAY)

        self._ready = False
        # Create Pub socket to broadcast to witnesses
        self.pub_sock = self.manager.create_socket(socket_type=zmq.PUB, name="TxBatcher-PUB", secure=True)
        self.pub_sock.bind(port=MN_TX_PUB_PORT, ip=self.ip)

        # Create DEALER socket to talk to the BlockManager process over IPC
        self.ipc_dealer = None
        self._create_dealer_ipc(port=ipc_port, ip=ipc_ip, identity=str(0).encode())

    def serve(self):
        self.tasks.append(self.compose_transactions())

        # Start main event loop
        # self.loop.run_until_complete(self.compose_transactions())
        self.loop.run_until_complete(asyncio.gather(*self.tasks))

    def _create_dealer_ipc(self, port: int, ip: str, identity: bytes):
        self.ipc_dealer = self.manager.create_socket(socket_type=zmq.DEALER, name="Batcher-IPC-Dealer[{}]".format(0), secure=False)
        self.ipc_dealer.setsockopt(zmq.IDENTITY, identity)

        self.tasks.append(self.ipc_dealer.add_handler(handler_func=self.handle_ipc_msg))

    def _connect_dealer_ipc(self):
        self.log.notice("Connecting to BA's ROUTER socket with a DEALER using"
                        "ip {}, port {}".format(self.ipc_ip, self.ipc_port))
        self.ipc_dealer.connect(port=self.ipc_port, protocol='ipc', ip=self.ipc_ip)

    def handle_ipc_msg(self, frames):
        assert len(frames) == 2, "Expected 2 frames: (msg_type, msg_blob). Got {} instead.".format(frames)

        msg_type = frames[0]
        msg_blob = frames[1]

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        self.log.info('Got message on IPC {}'.format(msg_type))

        if msg_type == MessageType.BURN_INPUT_HASHES:
            self.log.info('An empty or non-empty block was made.')
            self.batcher.remove_batch_ids(msg.inputHashes)

        elif msg_type == MessageType.READY:
            self.log.success('READY.')
            self._ready = True

        else:
            self.log.error("Batcher got unexpected message type {} from BA's IPC "
                           "socket. Ignoring the msg {}".format(type(msg), msg))

    async def _wait_until_ready(self):
        await asyncio.sleep(0)
        self._connect_dealer_ipc()
        while not self._ready:
            await asyncio.sleep(0)

    async def compose_transactions(self):
        await self._wait_until_ready()

        self.log.notice("Starting TransactionBatcher ...")

        enc_fltr = TRANSACTION_FILTER.encode()

        while True:
            mtype, msg = await self.batcher.get_next_batch_packed()
            self.pub_sock.send_msg(msg=msg, msg_type=mtype, filter=enc_fltr)
            self.log.info("Sent next batch ...")





from unittest import TestCase
from cilantro_ee.core.containers.linked_hashtable import LinkedHashTable


class TestLinkedHashTable(TestCase):

    def test_empty_append(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        lht.append(k, v)

        self.assertTrue(k in lht)
        self.assertEquals(lht._last.value, v)
        self.assertEquals(lht._first.value, v)
        self.assertEquals(len(lht), 1)

    def test_nonempty_append(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        k2, v2 = 'another key', b'a thiccer value'
        k3, v3 = 'anotha one!', b'a thiccest value'
        k4, v4 = 'yet another one', b'a thiccerest value'
        lht.append(k, v)
        lht.append(k2, v2)
        lht.append(k3, v3)
        lht.append(k4, v4)

        for _k in (k, k2, k3):
            self.assertTrue(_k in lht)

        self.assertEquals(lht._first.value, v)
        self.assertEquals(lht._last.value, v4)
        self.assertEquals(lht._first.next.value, v2)
        self.assertEquals(lht._last.previous.value, v3)
        self.assertEquals(len(lht), 4)

    def test_clear(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        k2, v2 = 'another key', b'a thiccer value'
        k3, v3 = 'anotha one!', b'a thiccest value'
        k4, v4 = 'yet another one', b'a thiccerest value'
        lht.append(k, v)
        lht.append(k2, v2)
        lht.append(k3, v3)
        lht.append(k4, v4)

        lht.clear()

        self.assertEquals(len(lht), 0)
        self.assertEquals(lht._last, None)
        self.assertEquals(lht._first, None)

    def test_pop_empty(self):
        lht = LinkedHashTable()
        self.assertEquals(None, lht.pop())

    def test_popleft_empty(self):
        lht = LinkedHashTable()
        self.assertEquals(None, lht.popleft())

    def test_pop(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        k2, v2 = 'another key', b'a thiccer value'
        k3, v3 = 'anotha one!', b'a thiccest value'
        k4, v4 = 'yet another one', b'a thiccerest value'
        lht.append(k, v)
        lht.append(k2, v2)
        lht.append(k3, v3)
        lht.append(k4, v4)

        item = lht.pop()

        self.assertEquals(item, v4)
        self.assertEquals(lht._last.value, v3)
        self.assertEquals(lht._first.value, v)
        self.assertEquals(len(lht), 3)

    def test_popleft(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        k2, v2 = 'another key', b'a thiccer value'
        k3, v3 = 'anotha one!', b'a thiccest value'
        k4, v4 = 'yet another one', b'a thiccerest value'
        lht.append(k, v)
        lht.append(k2, v2)
        lht.append(k3, v3)
        lht.append(k4, v4)

        item = lht.popleft()

        self.assertEquals(item, v)
        self.assertEquals(lht._first.value, v2)
        self.assertEquals(lht._last.value, v4)
        self.assertEquals(len(lht), 3)

    def test_remove_leftmost(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        k2, v2 = 'another key', b'a thiccer value'
        k3, v3 = 'anotha one!', b'a thiccest value'
        k4, v4 = 'yet another one', b'a thiccerest value'
        lht.append(k, v)
        lht.append(k2, v2)
        lht.append(k3, v3)
        lht.append(k4, v4)

        item = lht.remove(k)

        self.assertEquals(item, v)
        self.assertEquals(lht._first.value, v2)
        self.assertEquals(lht._last.value, v4)
        self.assertEquals(len(lht), 3)

    def test_remove_rightmost(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        k2, v2 = 'another key', b'a thiccer value'
        k3, v3 = 'anotha one!', b'a thiccest value'
        k4, v4 = 'yet another one', b'a thiccerest value'
        lht.append(k, v)
        lht.append(k2, v2)
        lht.append(k3, v3)
        lht.append(k4, v4)

        item = lht.remove(k4)

        self.assertEquals(item, v4)
        self.assertEquals(lht._last.value, v3)
        self.assertEquals(lht._first.value, v)
        self.assertEquals(len(lht), 3)

    def test_remove_inner(self):
        lht = LinkedHashTable()

        k, v = 'some key', b'a thicc value'
        k2, v2 = 'another key', b'a thiccer value'
        k3, v3 = 'anotha one!', b'a thiccest value'
        k4, v4 = 'yet another one', b'a thiccerest value'
        lht.append(k, v)
        lht.append(k2, v2)
        lht.append(k3, v3)
        lht.append(k4, v4)

        item = lht.remove(k2)

        self.assertEquals(item, v2)

        self.assertEquals(lht._last.value, v4)
        self.assertEquals(lht._first.value, v)

        self.assertEquals(lht._first.next.value, v3)
        self.assertEquals(lht._last.previous.value, v3)
        self.assertEquals(lht._last.previous.previous.value, v)
        self.assertEquals(len(lht), 3)




from unittest import TestCase
from cilantro_ee.nodes.masternode.transaction_batcher import RateLimitingBatcher
from cilantro_ee.core.crypto.wallet import Wallet, _verify
from cilantro_ee.services.storage.state import NonceManager
from cilantro_ee.core.utils.transaction import TransactionBuilder, transaction_is_valid
from contracting import config
from cilantro_ee.core.messages.capnp_impl import capnp_struct as schemas
import os
import capnp

transaction_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/transaction.capnp')

n = NonceManager()


class MockQueue:
    def __init__(self):
        self.q = []

    def get(self):
        return self.q.pop(0)

    def qsize(self):
        return len(self.q)


def make_good_tx(processor):
    w = Wallet()
    balances_key = '{}{}{}{}{}'.format('currency',
                                       config.INDEX_SEPARATOR,
                                       'balances',
                                       config.DELIMITER,
                                       w.verifying_key().hex())
    n.driver.set(balances_key, 500000)
    tx = TransactionBuilder(w.verifying_key(),
                            contract='currency',
                            function='transfer',
                            kwargs={'amount': 10, 'to': 'jeff'},
                            stamps=500000,
                            processor=processor,
                            nonce=0)

    tx.sign(w.signing_key())
    tx_bytes = tx.serialize()
    tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)
    return tx_struct


def make_bad_tx():
    w = Wallet()
    balances_key = '{}{}{}{}{}'.format('currency',
                                       config.INDEX_SEPARATOR,
                                       'balances',
                                       config.DELIMITER,
                                       w.verifying_key().hex())
    n.driver.set(balances_key, 500000)
    tx = TransactionBuilder(w.verifying_key(),
                            contract='currency',
                            function='transfer',
                            kwargs={'amount': 10, 'to': 'jeff'},
                            stamps=500000,
                            processor=b'\x00' * 32,
                            nonce=0)

    tx.sign(w.signing_key())
    tx_bytes = tx.serialize()
    tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)
    return tx_struct


class TestRateLimitingBatcher(TestCase):
    def test_add_batch_id_adds_properly(self):
        r = RateLimitingBatcher(
            queue=MockQueue(),
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=1,
            max_txn_delay=1
        )

        r.add_batch_id('123')

        self.assertListEqual(r.sent_batch_ids, ['123'])
        self.assertEqual(r.num_batches_sent, 1)

    def test_adding_multiple_batches_adjusts_accordingly(self):
        r = RateLimitingBatcher(
            queue=MockQueue(),
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=1,
            max_txn_delay=1
        )

        r.add_batch_id('123')

        self.assertListEqual(r.sent_batch_ids, ['123'])
        self.assertEqual(r.num_batches_sent, 1)

        r.add_batch_id('567')

        self.assertListEqual(r.sent_batch_ids, ['123', '567'])
        self.assertEqual(r.num_batches_sent, 2)

    def test_adding_same_batch_ids_acts_like(self):
        # TBD
        pass

    def test_ready_for_next_batch_batches_sent_greater_than_2_returns_false(self):
        queue = MockQueue()

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=1,
            max_txn_delay=1
        )

        r.num_batches_sent = 3

        self.assertFalse(r.ready_for_next_batch())
        self.assertEqual(r.txn_delay, 0)

    def test_ready_for_next_batch_batches_sent_greater_than_2_returns_false_and_increments_if_tx_exists(self):
        queue = MockQueue()
        queue.q.append('tx!')

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=1,
            max_txn_delay=1
        )

        r.num_batches_sent = 3

        self.assertFalse(r.ready_for_next_batch())
        self.assertEqual(r.txn_delay, 1)

    def test_ready_for_next_batch_less_than_2_num_batches_sent_1_q_greater_than_max_batches_returns_false_and_increments(self):
        queue = MockQueue()

        queue.q.extend(['1', '2'])

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.num_batches_sent = 1

        self.assertFalse(r.ready_for_next_batch())
        self.assertEqual(r.txn_delay, 1)

    def test_ready_for_next_batch_lt_2_num_batches_tx_delay_lt_max_txn_submission_delay_returns_false(self):
        queue = MockQueue()

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.num_batches_sent = 1

        self.assertFalse(r.ready_for_next_batch())
        self.assertEqual(r.txn_delay, 0)

    def test_ready_for_next_batch_lt_2_num_batches_tx_delay_lt_max_txn_submission_delay_returns_false_and_increments(self):
        queue = MockQueue()
        queue.q.extend(['1', '2'])

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.num_batches_sent = 1

        self.assertFalse(r.ready_for_next_batch())
        self.assertEqual(r.txn_delay, 1)

    def test_num_batches_lt_2_num_batches_sent_gt_0_and_q_gt_mx_batches_and_tx_delay_gt_max_delay_returns_true(self):
        queue = MockQueue()
        queue.q.extend([1, 2, 3, 4, 5, 6, 7])

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.num_batches_sent = 1
        r.txn_delay = 100

        self.assertTrue(r.ready_for_next_batch())
        self.assertEqual(r.txn_delay, 101)

    def test_remove_batch_ids_removes_id(self):
        queue = MockQueue()

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.add_batch_id('1')
        r.add_batch_id('2')
        r.add_batch_id('3')

        r.remove_batch_ids(['1', '2'])

        self.assertListEqual(r.sent_batch_ids, ['3'])

    def test_remove_batch_ids_decrements_num_batches_sent(self):
        queue = MockQueue()

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.add_batch_id('1')
        r.add_batch_id('2')
        r.add_batch_id('3')

        self.assertEqual(r.num_batches_sent, 3)

        r.remove_batch_ids(['1', '2'])

        self.assertEqual(r.num_batches_sent, 1)

    def test_remove_batch_ids_sets_sent_batch_ids_to_new_id_list_if_corruption_occurs(self):
        queue = MockQueue()

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.num_batches_sent = 1000

        r.add_batch_id('1')
        r.add_batch_id('2')
        r.add_batch_id('3')

        r.remove_batch_ids(['1', '2'])

        self.assertEqual(r.num_batches_sent, 1)

    def test_get_next_batch_size_num_tx_gte_max_batch_size_returns_smaller_size(self):
        queue = MockQueue()

        queue.q = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        self.assertEqual(r.get_next_batch_size(), 3)

    def test_get_next_batch_size_num_tx_gte_max_batch_size_resets_txn_delay(self):
        queue = MockQueue()

        queue.q = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1000000
        )

        r.txn_delay = 10000

        r.get_next_batch_size()

        self.assertEqual(r.txn_delay, 0)

    def test_get_next_batch_size_txn_delay_gte_max_returns_smaller_size(self):
        queue = MockQueue()

        queue.q = [1]

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )
        r.txn_delay = 10000
        self.assertEqual(r.get_next_batch_size(), 1)

    def test_get_next_batch_size_txn_delay_gte_max_returns_resets_txn_delay(self):
        queue = MockQueue()

        queue.q = [1]

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        r.txn_delay = 10000
        r.get_next_batch_size()
        self.assertEqual(r.txn_delay, 0)

    def test_get_next_batch_num_txn_lt_max_batch_txn_delay_lt_max_returns_0(self):
        queue = MockQueue()

        queue.q = [1]

        r = RateLimitingBatcher(
            queue=queue,
            wallet=Wallet(),
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        self.assertEqual(r.get_next_batch_size(), 0)

    def test_get_txn_list_filters_bad_txs(self):
        w = Wallet()
        queue = MockQueue()

        r = RateLimitingBatcher(
            queue=queue,
            wallet=w,
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        a = make_good_tx(w.vk.encode())
        b = make_bad_tx()
        c = make_bad_tx()
        d = make_good_tx(w.vk.encode())

        expected = [a, d]

        queue.q.extend([(1, a), (1, b), (1, c), (1, d)])

        self.assertListEqual(r.get_txn_list(4), expected)

    def test_get_txn_list_all_if_all_good(self):
        w = Wallet()
        queue = MockQueue()

        r = RateLimitingBatcher(
            queue=queue,
            wallet=w,
            sleep_interval=0,
            max_batch_size=3,
            max_txn_delay=1
        )

        a = make_good_tx(w.vk.encode())
        b = make_good_tx(w.vk.encode())
        c = make_good_tx(w.vk.encode())
        d = make_good_tx(w.vk.encode())

        expected = [a, b, c, d]

        queue.q.extend([(1, a), (1, b), (1, c), (1, d)])

        self.assertListEqual(r.get_txn_list(4), expected)




supply = Variable()
balances = Hash(default_value=0)

@construct
def seed():
    #balances[ctx.caller] = 288_090_567

    seed_amount = 1000000
    supply.set(0)

    founder_wallets = [
        "8143f569f258c33de00d71a6527eadd5e22a88b0f16bee3f13230d26dd8a5e46",
        "6fa435ea82e54d55b20599609b750f2602c596edbd70520c156a7d907e26926b",
        "9dcef7bceb11e0d7f4ce687d343be819fb96c4abc3bb87fe4126849230c93a2a",
        "64cdc0774705e09d1f2c3288766721ad1fa4a9cfd2be8260c127c0ea3475dd8e",
        "772de9fcb7a44d4d91c6419f9f79cdd81b9ba874400b57365b288fac20d28b2c",
        "1f2995f6cb7afad770efdf4615be264fae145d796fc7fb3b9794372812e361cd",
        "2b060b761c8e698168cc9641ef2b22ff357464cb324866631572508defcc0898",
        "ed3722a66e806935c58f830813a9eb69f1beadff704ad1e611b741d0c7fb6190",
        "b8491def67aa7f7bf03d1d2642e029ef82f6c73e8478e78ebe536e3c2b011aec",
        "76bebe49e768818773bda994491dc271f8e360917a820278d4e52d7a7f76c9af",
        "d0907067177ad93e6cb27df90e2c0c48d9f881ce5b4a00d51c09ef5dbe4d90a6",
        "41250f9159790f9c9b9e5cd558b2e42a86836b538dd091f2c747998fc90fa257",
        "e5c2815253b301f50a4e63ce7c5c61db9c65f543333e15257c34cf5c9f0b46e5",
        "8e16d6c9054cc90b4f7443ae3ceeee3fd64db22eb46f2421c51bc5f203ecbb15",
        "1f328c2ff4ccdfe1b66724304cc88be3b00fd08d85851a086440d4faf0858888",
        "93aeefec5d87bc890b02fec280ab9c79803d2ab71d1cb274d7090e9b84cd6f8a",
        "3c820d80eadfef392ab97ceb9c0d3459f45d14be5cbcb680f84910d7778483ae",
        "aacfe4211a363363211f300a9bfcd70072d2615ccf58cd7fdb53c2a8d74d1c7e",
        "dcbcb40c83e865f805e30842b810c330031645edc405ebe6376fda84da2d82b6",
        "ec6361c4849faad114748b00a801076b0ddbeda78164c51079ab0abab5f2b6bd",
        "7cc186a08d7a2f510d2670af4787b73b40d91abd8d1817125f0de850f2e9bd96",
        "5d7f99022d87d344150b598e79477a580ae1ccd8d681ba46330597dbbdfd91bc",
        "db3ea9f7b716896bae3efa78da381e81fc6707c1a2387ffd3e97e62b00b2bc69",
        "f022edef56754196edfb591ea9f5878164fd9ce26e5c4fc2ee9260717cd751a7",
        "5568d257e567c043e7fbf8d6425256a174da2719f6829bbe46a4c2bbfc8bb9a4",
        "1bf8eae89e98dcd8066207a667c30a4d6345886c148eba359f0a95fe701f754d",
    ]

    for w in founder_wallets:
        # print('Minting {} with {} coins'.format(w, seed_amount))
        balances[w] = seed_amount

        s = supply.get()
        s += seed_amount
        supply.set(s)
    # print('Done minting!')

@export
def transfer(amount, to):
    sender = ctx.caller

    balance = balances[sender]

    assert balance >= amount, 'Current balance {} from sender {} is less than amount {}.'.format(
        balance,
        sender,
        amount
    )

    balances[sender] -= amount
    balances[to] += amount

    # print('Succesfully sent {} coins to {}. {} updated from {} to {}'.format(
        # amount,
        # to,
        # sender,
        # balance,
        # balances[sender]
    # ))

@export
def balance_of(account):
    return balances[account]

@export
def total_supply():
    return supply.get()

@export
def allowance(owner, spender):
    return balances[owner, spender]

@export
def approve(amount, to):
    sender = ctx.caller
    balances[sender, to] += amount
    return balances[sender, to]

@export
def transfer_from(amount, to, main_account):
    sender = ctx.caller

    assert balances[main_account, sender] >= amount, 'Not enough coins approved to send! You have {} and are trying to spend {}'\
        .format(balances[main_account, sender], amount)
    assert balances[main_account] >= amount, 'Not enough coins to send!'

    balances[main_account, sender] -= amount
    balances[main_account] -= amount

    balances[to] += amount



from unittest import TestCase
from cilantro_ee.core.sockets.socket_book import SocketBook
from cilantro_ee.services.storage.vkbook import PhoneBook, VKBook
from cilantro_ee.services.overlay.network import Network
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.sockets.services import SocketStruct, _socket

import zmq
import zmq.asyncio

import asyncio


class TestSocketBook(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.ctx.destroy()
        self.loop.close()

    def test_new_nodes(self):
        current_nodes = {1, 2, 3, 4}
        all_nodes = {1, 2, 3, 4, 5, 6, 7, 8}

        self.assertEqual(SocketBook.new_nodes(all_nodes, current_nodes), {5, 6, 7, 8})

    def test_old_nodes(self):
        current_nodes = {1, 2, 3, 4}
        all_nodes = {1, 2, 3, 4, 5, 6, 7, 8}

        self.assertEqual(SocketBook.old_nodes(all_nodes, current_nodes), set())

    def test_old_nodes_actual_difference(self):
        current_nodes = {1, 2, 3, 4}
        all_nodes = {1, 4, 5, 6, 7, 8}

        self.assertEqual(SocketBook.old_nodes(all_nodes, current_nodes), {2, 3})

    def test_remove_node(self):
        m = SocketBook()
        ctx = zmq.Context()
        m.sockets = {'a': ctx.socket(zmq.PUB), 'b': ctx.socket(zmq.PUB)}

        m.remove_node('a')

        self.assertIsNone(m.sockets.get('a'))
        self.assertIsNotNone(m.sockets.get('b'))

    def test_remove_node_doesnt_exist_does_nothing(self):
        m = SocketBook()
        ctx = zmq.Context()
        m.sockets = {'a': ctx.socket(zmq.PUB), 'b': ctx.socket(zmq.PUB)}

        m.remove_node('c')

        self.assertIsNotNone(m.sockets.get('a'))
        self.assertIsNotNone(m.sockets.get('b'))

    def test_refresh(self):
        PhoneBook = VKBook(masternodes=['stu', 'raghu'],
                           delegates=['tejas', 'alex', 'steve'],
                           num_boot_mns=2,
                           num_boot_del=3,
                           stamps=True,
                           nonces=True)

        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        expected = {
            'stu': '127.0.0.1',
            'raghu': '127.0.0.2'
        }
        p1.peer_service.table.peers = expected
        masternodes = SocketBook(network=p1, phonebook_function=PhoneBook.contract.get_masternodes)

        self.assertDictEqual(masternodes.sockets, {})

        loop = asyncio.get_event_loop()
        loop.run_until_complete(masternodes.refresh())

        self.assertDictEqual(masternodes.sockets, expected)

    def test_refresh_remove_old_nodes(self):
        PhoneBook = VKBook(masternodes=['stu', 'raghu'],
                           delegates=['tejas', 'alex', 'steve'],
                           num_boot_mns=2,
                           num_boot_del=3,
                           stamps=True,
                           nonces=True)

        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        ctx = zmq.Context()

        peeps = {
            'stu': ctx.socket(zmq.SUB),
            'raghu': ctx.socket(zmq.SUB),
            'tejas': ctx.socket(zmq.SUB),
            'steve': ctx.socket(zmq.SUB)
        }

        p1.peer_service.table.peers = peeps
        masternodes = SocketBook(network=p1, phonebook_function=PhoneBook.contract.get_masternodes)

        self.assertDictEqual(masternodes.sockets, {})

        loop = asyncio.get_event_loop()
        loop.run_until_complete(masternodes.refresh())

        expected = {
            'stu': peeps['stu'],
            'raghu': peeps['raghu'],
        }

        self.assertDictEqual(masternodes.sockets, expected)

        PhoneBook = VKBook(masternodes=['stu', 'tejas'],
                           delegates=['raghu', 'alex', 'steve'],
                           num_boot_mns=2,
                           num_boot_del=3,
                           stamps=True,
                           nonces=True)

        loop = asyncio.get_event_loop()
        loop.run_until_complete(masternodes.refresh())

        expected = {
            'stu': peeps['stu'],
            'tejas': peeps['tejas']
        }

        self.assertDictEqual(masternodes.sockets, expected)



from unittest import TestCase
from unittest.mock import MagicMock

from cilantro_ee.messages.transaction.contract import ContractTransaction, ContractTransactionBuilder
from cilantro_ee.messages.transaction.data import TransactionData, TransactionDataBuilder
from cilantro_ee.constants.testnet import TESTNET_MASTERNODES
TEST_SK = TESTNET_MASTERNODES[0]['sk']


class TestTransactionData(TestCase):

    def test_create_with_contract_tx(self):
        tx = ContractTransactionBuilder.random_currency_tx()
        status = 'SUCC'
        state = 'SET GHU 10;'

        tx_data = TransactionData.create(contract_tx=tx, status=status, state=state)

        self.assertEqual(tx_data.transaction, tx)
        self.assertEqual(tx_data.status, status)
        self.assertEqual(tx_data.state, state)

    def test_serialize_deserialize(self):
        tx = ContractTransactionBuilder.random_currency_tx()
        status = 'SUCC'
        state = 'SET GHU 10;'

        tx_data = TransactionData.create(contract_tx=tx, status=status, state=state)
        clone = TransactionData.from_bytes(tx_data.serialize())

        self.assertEqual(tx_data, clone)

    def test_create(self):
        td = TransactionData.create(
            contract_tx=ContractTransactionBuilder.create_currency_tx(
                sender_sk=TEST_SK, receiver_vk='A' * 64, amount=10),
            status='SUCCESS', state='SET x 1')

        self.assertTrue(isinstance(td.transaction, ContractTransaction))

    def test_serialize_deserialize2(self):
        td = TransactionData.create(
            contract_tx=ContractTransactionBuilder.create_currency_tx(
                sender_sk=TEST_SK, receiver_vk='A' * 64, amount=10),
            status='SUCCESS', state='SET x 1')
        clone = TransactionData.from_bytes(td.serialize())

        self.assertEqual(clone, td)

    def test_serialize_deserialize_consistent_hash(self):
        td = TransactionData.create(
            contract_tx=ContractTransactionBuilder.create_currency_tx(
                sender_sk=TEST_SK, receiver_vk='A' * 64, amount=10),
            status='SUCCESS', state='SET x 1')
        clone = TransactionData.from_bytes(td.serialize())

        self.assertEqual(clone.hash, td.hash)



from cilantro_ee.core.sockets.services import AsyncInbox

from cilantro_ee.services.storage.master import CilantroStorageDriver
from cilantro_ee.core.top import TopBlockManager
from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.messages.message_type import MessageType
import os
import capnp
from cilantro_ee.core.messages.capnp_impl import capnp_struct as schemas
import json
import struct
import time

blockdata_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/blockdata.capnp')
subblock_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/subblock.capnp')


# Provide a block blocks to enable data and index requests
# Otherwise, this will just return latest num and hash, which both delegates and masters can do
class BlockServer(AsyncInbox):
    def __init__(self, socket_id, wallet, ctx, linger, poll_timeout, driver: CilantroStorageDriver=None, top=TopBlockManager()):
        self.driver = driver
        self.top = top
        super().__init__(socket_id=socket_id,
                         wallet=wallet,
                         ctx=ctx,
                         linger=linger,
                         poll_timeout=poll_timeout)

    async def handle_msg(self, _id, msg):
        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(message=msg)

        if msg_type == MessageType.BLOCK_DATA_REQUEST and self.driver is not None:

            block_dict = self.driver.get_block(msg.blockNum)

            if block_dict is not None:
                block_hash = block_dict.get('blockHash')
                block_num = block_dict.get('blockNum')
                prev_hash = block_dict.get('prevBlockHash')
                subblocks = block_dict.get('subBlocks')
                owners = block_dict.get('blockOwners')

                reply = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                            msg_type=MessageType.BLOCK_DATA,
                                                            blockHash=block_hash,
                                                            blockNum=block_num,
                                                            blockOwners=[owner for owner in owners],
                                                            prevBlockHash=prev_hash,
                                                            subBlocks=[subblock_capnp.SubBlock.new_message(**sb) for sb in subblocks],
                                                            )

                await self.return_msg(_id, reply)
            else:
                reply = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                            msg_type=MessageType.BAD_REQUEST,
                                                            timestamp=int(time.time()))
                await self.return_msg(_id, reply)

        elif msg_type == MessageType.BLOCK_INDEX_REQUEST and self.driver is not None:
            await self.return_msg(_id, b'howdy')

        elif msg_type == MessageType.LATEST_BLOCK_HEIGHT_REQUEST:
            reply = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                        msg_type=MessageType.LATEST_BLOCK_HEIGHT_REPLY,
                                                        blockHeight=self.top.get_latest_block_number())

            await self.return_msg(_id, reply)

        elif msg_type == MessageType.LATEST_BLOCK_HASH_REQUEST:
            reply = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                        msg_type=MessageType.LATEST_BLOCK_HASH_REPLY,
                                                        blockHash=self.top.get_latest_block_hash())

            await self.return_msg(_id, reply)
        else:
            reply = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                        msg_type=MessageType.BAD_REQUEST,
                                                        timestamp=int(time.time()))
            await self.return_msg(_id, reply)



from os import getenv as env

ALPHA = 2
KSIZE = 4
MAX_PEERS = 64

AUTH_TIMEOUT = 20
RPC_TIMEOUT = 10
FIND_NODE_HOP_TIMEOUT = 10
FIND_NODE_TIMEOUT = FIND_NODE_HOP_TIMEOUT * 4   # we should multiply by log2(network size)

# The number of discovery retries necessary before a masternode boots alone
DISCOVERY_WAIT = 3
DISCOVERY_RETRIES = 3
DISCOVERY_LONG_WAIT = 10
DISCOVERY_ITER = 50

# How long OverlayClient should wait for a rdy sig from the OverlayServer until we timeout
# CLIENT_SETUP_TIMEOUT = DISCOVERY_LONG_WAIT * DISCOVERY_ITER
CLIENT_SETUP_TIMEOUT = 10          # fixed amount in EE version

PEPPER = env('PEPPER', 'cilantro_pepper')
EVENT_URL = 'ipc://overlay-event-ipc-sock'
CMD_URL = 'ipc://overlay-cmd-ipc-sock'



from unittest import TestCase
from cilantro_ee.services.storage.master import MasterStorage


class TestMasterStorage(TestCase):
    def setUp(self):
        self.db = MasterStorage()

    def tearDown(self):
        self.db.drop_collections()

    def test_init(self):
        self.assertTrue(self.db)

    def test_q_num(self):
        q = self.db.q(1)

        self.assertEqual(q, {'blockNum': 1})

    def test_q_hash(self):
        q = self.db.q('1')

        self.assertEqual(q, {'blockHash': '1'})

    def test_genesis_block_created(self):
        block = self.db.get_block(0)

        expected = {
                'blockNum': 0,
                'blockHash': '0' * 64,
                'blockOwners': ['0' * 64]
            }

        self.assertEqual(block, expected)

        index = self.db.get_index(0)

        self.assertEqual(index, expected)

    def test_put_block(self):
        block = {
            'blockHash': 'a',
            'blockNum': 1,
            'data': 'woop'
        }

        _id = self.db.put(block)

        self.assertTrue(_id)

    def test_get_block(self):
        block = {
            'blockHash': 'a',
            'blockNum': 1,
            'data': 'woop'
        }

        _id = self.db.put(block)

        self.assertTrue(_id)

        got_block = self.db.get_block(1)

        block.pop('_id')

        self.assertEqual(block, got_block)

    def test_get_block_hash(self):
        block = {
            'blockHash': 'a',
            'blockNum': 1,
            'data': 'woop'
        }

        _id = self.db.put(block)

        self.assertTrue(_id)

        got_block = self.db.get_block('a')

        block.pop('_id')

        self.assertEqual(block, got_block)

    def test_get_none_block(self):
        block = {
            'blockHash': 'a',
            'blockNum': 1,
            'data': 'woop'
        }

        _id = self.db.put(block)

        self.assertTrue(_id)

        got_block = self.db.get_block('b')

        block.pop('_id')

        self.assertIsNone(got_block)

    def test_got_none_block_num(self):
        block = {
            'blockHash': 'a',
            'blockNum': 1,
            'data': 'woop'
        }

        _id = self.db.put(block)

        self.assertTrue(_id)

        got_block = self.db.get_block(2)

        block.pop('_id')

        self.assertIsNone(got_block)

    def test_drop_collections_block(self):
        block = {
            'blockHash': 'a',
            'blockNum': 1,
            'data': 'woop'
        }

        _id = self.db.put(block)

        self.assertTrue(_id)

        self.db.drop_collections()

        got_block = self.db.get_block(1)

        block.pop('_id')

        self.assertIsNone(got_block)

    def test_put_index(self):
        index = {
            'blockHash': 'a',
            'blockNum': 1,
            'blockOwners': 'stu'
        }

        _id = self.db.put(index, MasterStorage.INDEX)

        self.assertTrue(_id)

    def test_put_other(self):
        index = {
            'blockHash': 'a',
            'blockNum': 1,
            'blockOwners': 'stu'
        }

        _id = self.db.put(index, 999)

        self.assertFalse(_id)

    def test_get_owners_num(self):
        index = {
            'blockHash': 'a',
            'blockNum': 1,
            'blockOwners': 'stu'
        }

        _id = self.db.put(index, MasterStorage.INDEX)

        self.assertTrue(_id)

        owners = self.db.get_owners(1)

        self.assertEqual(owners, 'stu')

    def test_get_owners_hash(self):
        index = {
            'blockHash': 'a',
            'blockNum': 1,
            'blockOwners': 'stu'
        }

        _id = self.db.put(index, MasterStorage.INDEX)

        self.assertTrue(_id)

        owners = self.db.get_owners('a')

        self.assertEqual(owners, 'stu')

    def test_get_owners_doesnt_exist(self):
        index = {
            'blockHash': 'a',
            'blockNum': 1,
            'blockOwners': 'stu'
        }

        _id = self.db.put(index, MasterStorage.INDEX)

        self.assertTrue(_id)

        owners = self.db.get_owners('b')

        self.assertIsNone(owners)

    def test_get_last_n_blocks(self):
        blocks = []

        blocks.append({'blockHash': 'a', 'blockNum': 1, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 2, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 3, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 4, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 5, 'data': 'woop'})

        for block in blocks:
            self.db.put(block)

        got_blocks = self.db.get_last_n(3, MasterStorage.BLOCK)

        nums = [b['blockNum'] for b in got_blocks]

        self.assertEqual(nums, [5, 4, 3])

    def test_get_last_n_index(self):
        blocks = []

        blocks.append({'blockHash': 'a', 'blockNum': 1, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 2, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 3, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 4, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 5, 'data': 'woop'})

        for block in blocks:
            self.db.put(block, MasterStorage.INDEX)

        got_blocks = self.db.get_last_n(3, MasterStorage.INDEX)

        nums = [b['blockNum'] for b in got_blocks]

        self.assertEqual(nums, [5, 4, 3])

    def test_get_none_from_wrong_n_collection(self):
        blocks = []

        blocks.append({'blockHash': 'a', 'blockNum': 1, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 2, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 3, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 4, 'data': 'woop'})
        blocks.append({'blockHash': 'a', 'blockNum': 5, 'data': 'woop'})

        for block in blocks:
            self.db.put(block, MasterStorage.INDEX)

        got_blocks = self.db.get_last_n(3, 5)

        self.assertIsNone(got_blocks)



from unittest import TestCase
from cilantro_ee.services.overlay.client import OverlayClient
import zmq
import zmq.asyncio
import asyncio


async def stop_server(s, timeout):
    await asyncio.sleep(timeout)
    s.stop()


class TestOverlayClient(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.ctx.destroy()
        self.loop.close()

    def test_sub_socket_receives_events(self):
        def reply_callback(*args, **kwargs):
            pass

        l = []

        def event_callback(*args, **kwargs):
            l.append('got event!')

        async def send_event():
            self.evt_sock.send_json({'hello': 'there'})

        o = OverlayClient(reply_callback, event_callback, self.ctx)

        self.evt_sock = self.ctx.socket(socket_type=zmq.PUB)
        self.evt_sock.bind('tcp://127.0.0.1:10003')

        tasks = asyncio.gather(
            o.start(),
            send_event(),
            stop_server(o, 0.1)
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)



from unittest import TestCase
from cilantro_ee.utils import Hasher
from cilantro_ee.core.containers.merkle_tree import MerkleTree
import secrets


class TestMerkleTree(TestCase):
    def test_construction_1(self):
        should_be = [None, None, None, 1, 2, 3, 4]
        m = MerkleTree([1, 2, 3, 4])
        self.assertEqual(len(should_be), len(m.nodes))

    def test_construction_2(self):
        should_be = [None, None, 1, 2, 3]
        m = MerkleTree([1, 2, 3])
        self.assertEqual(len(should_be), len(m.nodes))

    def test_construction_3(self):
        should_be = [None, None, None, None, None, 1, 2, 3, 4, 5, 6]
        m = MerkleTree([1, 2, 3, 4, 5, 6])
        self.assertEqual(len(should_be), len(m.nodes))

    def test_make_merkle_works(self):
        leaves = [1, 2, 3, 4]
        test_merkle = [None,
                       None,
                       None,
                       MerkleTree.hash(1),
                       MerkleTree.hash(2),
                       MerkleTree.hash(3),
                       MerkleTree.hash(4)]

        test_merkle[2] = MerkleTree.hash(test_merkle[-2] + test_merkle[-1])
        test_merkle[1] = MerkleTree.hash(test_merkle[3] + test_merkle[4])
        test_merkle[0] = MerkleTree.hash(test_merkle[1] + test_merkle[2])

        m = MerkleTree(leaves)

        self.assertEqual(test_merkle, m.nodes)

    def test_find_root(self):
        leaves = [1, 2, 3, 4]
        test_merkle = [None,
                       None,
                       None,
                       MerkleTree.hash(1),
                       MerkleTree.hash(2),
                       MerkleTree.hash(3),
                       MerkleTree.hash(4)]

        test_merkle[2] = MerkleTree.hash(test_merkle[-2] + test_merkle[-1])
        test_merkle[1] = MerkleTree.hash(test_merkle[3] + test_merkle[4])
        test_merkle[0] = MerkleTree.hash(test_merkle[1] + test_merkle[2])

        m = MerkleTree(leaves)

        self.assertEqual(m.parent(2), test_merkle[0])
        self.assertEqual(m.parent(4), test_merkle[1])
        self.assertEqual(m.parent(), test_merkle[0])
        self.assertEqual(m.root, test_merkle[0])

    def test_find_children(self):
        leaves = [1, 2, 3, 4]
        test_merkle = [None,
                       None,
                       None,
                       MerkleTree.hash(1),
                       MerkleTree.hash(2),
                       MerkleTree.hash(3),
                       MerkleTree.hash(4)]

        test_merkle[2] = MerkleTree.hash(test_merkle[-2] + test_merkle[-1])
        test_merkle[1] = MerkleTree.hash(test_merkle[3] + test_merkle[4])
        test_merkle[0] = MerkleTree.hash(test_merkle[1] + test_merkle[2])

        m = MerkleTree(leaves)

        self.assertEqual(m.children(0), [test_merkle[1], test_merkle[2]])
        self.assertEqual(m.children(1), [test_merkle[3], test_merkle[4]])
        self.assertEqual(m.children(2), [test_merkle[5], test_merkle[6]])

    def test_verify_tree_from_bytes(self):
        raw_leaves = [secrets.token_bytes(16) for _ in range(16)]

        m = MerkleTree(raw_leaves)

        self.assertTrue(MerkleTree.verify_tree_from_bytes(m.leaves, m.root))

    def test_verify_tree_from_concat_str(self):
        raw_leaves = [secrets.token_bytes(16) for _ in range(16)]

        m = MerkleTree(raw_leaves)

        self.assertTrue(MerkleTree.verify_tree_from_concat_str(m.leaves_as_concat_hex_str, m.root_as_hex))

    def test_verify_tree_from_concat_str_invalid(self):
        raw_leaves = [secrets.token_bytes(16) for _ in range(16)]
        random_hash = MerkleTree.hash(secrets.token_bytes(16))

        m = MerkleTree(raw_leaves)

        self.assertFalse(MerkleTree.verify_tree_from_concat_str(m.leaves_as_concat_hex_str, random_hash))

    def test_verify_tree_from_bytes_invalid(self):
        raw_leaves = [secrets.token_bytes(16) for _ in range(16)]

        m = MerkleTree(raw_leaves)

        # Set the last leaf to a random value obviously different then whatever was used to build the tree
        # this should cause validation to fail
        bad_leaves = m.leaves
        bad_leaves[-1] = secrets.token_bytes(16)

        self.assertFalse(MerkleTree.verify_tree_from_bytes(bad_leaves, m.root))

    def test_leaves_as_hex(self):
        leaves = [1, 2, 3, 4]
        hashed_leaves = [MerkleTree.hash(leaf).hex() for leaf in leaves]

        m = MerkleTree(leaves)

        for actual, expected in zip(m.leaves_as_hex, hashed_leaves):
            self.assertEqual(actual, expected)

    def test_root_as_hex(self):
        leaves = [1, 2, 3, 4]
        test_merkle = [None,
                       None,
                       None,
                       MerkleTree.hash(1),
                       MerkleTree.hash(2),
                       MerkleTree.hash(3),
                       MerkleTree.hash(4)]

        test_merkle[2] = MerkleTree.hash(test_merkle[-2] + test_merkle[-1])
        test_merkle[1] = MerkleTree.hash(test_merkle[3] + test_merkle[4])
        test_merkle[0] = MerkleTree.hash(test_merkle[1] + test_merkle[2])

        m = MerkleTree(leaves)

        self.assertEqual(m.root, test_merkle[0])
        self.assertEqual(m.root_as_hex, test_merkle[0].hex())

    def test_hash_leaves_false(self):
        leaves = [1, 2, 3, 4]
        prehashed_leaves = list(map(MerkleTree.hash, leaves))

        merk_from_leaves = MerkleTree(leaves=leaves)
        merk_from_hashes = MerkleTree(leaves=prehashed_leaves, hash_leaves=False)

        for node1, node2 in zip(merk_from_hashes.nodes, merk_from_leaves.nodes):
            self.assertEqual(node1, node2)

    def test_leaves_as_hex_str(self):
        leaves = [1, 2, 3, 4]
        hashed_leaves = [MerkleTree.hash(leaf).hex() for leaf in leaves]
        hex_leaves = ''.join(hashed_leaves)

        m = MerkleTree(leaves)

        self.assertEqual(hex_leaves, m.leaves_as_concat_hex_str)

    def test_from_leaves_hex_str(self):
        leaves = [1, 2, 3, 4]

        m_from_init = MerkleTree(leaves)
        m_from_hex_str = MerkleTree.from_leaves_hex_str(m_from_init.leaves_as_concat_hex_str)

        self.assertEqual(m_from_hex_str.root, m_from_init.root)

    def test_data_for_hash(self):
        leaves = [b'1', b'2', b'3']
        hashed_leaves = list(map(Hasher.hash, leaves))

        m = MerkleTree.from_raw_transactions(leaves)

        for i in range(len(leaves)):
            self.assertEquals(m.data_for_hash(hashed_leaves[i]), leaves[i])

    def test_data_for_hash_doesnt_exist(self):
        leaves = [b'1', b'2', b'3']
        hashed_leaves = list(map(Hasher.hash, leaves))

        m = MerkleTree.from_raw_transactions(leaves)

        self.assertRaises(AssertionError, m.data_for_hash, '0' * 64)

    def test_very_tree_from_str(self):
        leaves = ['A' * 64, 'B' * 64, 'C' * 64]
        tree = MerkleTree.from_hex_leaves(leaves)
        root = tree.root_as_hex

        self.assertTrue(MerkleTree.verify_tree_from_str(leaves, root))

    def test_very_tree_from_str_invalid(self):
        leaves = ['A' * 64, 'B' * 64, 'C' * 64]
        tree = MerkleTree.from_hex_leaves(leaves)
        bad_root = 'DEADBEEF' * 8

        self.assertFalse(MerkleTree.verify_tree_from_str(leaves, bad_root))

    def test_from_raw_leaves_equals_from_hex_leaves(self):
        raw_bytes = [secrets.token_bytes(8) for _ in range(20)]

        tree_from_bytes = MerkleTree.from_raw_transactions(raw_bytes)
        tree_from_hex = MerkleTree.from_hex_leaves(tree_from_bytes.leaves_as_hex)

        self.assertEqual(tree_from_bytes.root, tree_from_hex.root)
        self.assertEqual(tree_from_bytes.root_as_hex, tree_from_hex.root_as_hex)

    def test_from_raw_leaves_equals_from_concat_hex_leaves(self):
        raw_bytes = [secrets.token_bytes(8) for _ in range(20)]

        tree_from_bytes = MerkleTree.from_raw_transactions(raw_bytes)
        tree_from_hex = MerkleTree.from_leaves_hex_str(tree_from_bytes.leaves_as_concat_hex_str)

        self.assertEqual(tree_from_bytes.root, tree_from_hex.root)
        self.assertEqual(tree_from_bytes.root_as_hex, tree_from_hex.root_as_hex)

    def test_from_hex_leaves_same(self):
        raw_bytes = [secrets.token_bytes(8) for _ in range(20)]

        tree_from_bytes = MerkleTree.from_raw_transactions(raw_bytes)
        tree_from_hex = MerkleTree.from_hex_leaves(tree_from_bytes.leaves_as_hex)
        other_tree_from_hex = MerkleTree.from_hex_leaves(tree_from_hex.leaves_as_hex)

        self.assertEqual(other_tree_from_hex.root, tree_from_hex.root)
        self.assertEqual(other_tree_from_hex.root_as_hex, tree_from_hex.root_as_hex)

    def test_from_hex_leaves(self):
        raw_bytes = [secrets.token_bytes(8) for _ in range(20)]
        hashes = [Hasher.hash(b, algorithm=MerkleTree.HASH_ALG) for b in raw_bytes]

        tree_from_bytes = MerkleTree.from_raw_transactions(raw_bytes)
        tree_from_str_leaves = MerkleTree.from_hex_leaves(hashes)

        self.assertEqual(tree_from_bytes.root, tree_from_str_leaves.root)
        self.assertEqual(tree_from_bytes.root_as_hex, tree_from_str_leaves.root_as_hex)

    def test_from_raw_leaves(self):
        raw_bytes = [secrets.token_bytes(8) for _ in range(20)]
        raw_hashes = [Hasher.hash(b, algorithm=MerkleTree.HASH_ALG, return_bytes=True) for b in raw_bytes]

        tree_from_raw_leaves = MerkleTree.from_raw_leaves(raw_hashes)
        tree_from_bytes = MerkleTree.from_raw_transactions(raw_bytes)

        self.assertEqual(tree_from_raw_leaves.root, tree_from_bytes.root)
        self.assertEqual(tree_from_raw_leaves.root_as_hex, tree_from_bytes.root_as_hex)





from unittest import TestCase
from cilantro_ee.core.crypto import zbase


class TestZBase(TestCase):
    def test_bytes_to_zbase_works(self):
        crypt = zbase.bytes_to_zbase32(b'hello123xx')
        self.assertEqual(crypt, 'pb1sa5dxgr3dg6da')

    def test_zbase_to_bytes_works(self):
        decrypt = zbase.zbase32_to_bytes('ctbs63dfrb7n4aubqp114c31')
        self.assertEqual(decrypt, b'dCode z-base-32')



from contracting.db.driver import ContractDriver
from cilantro_ee.core.messages.capnp_impl import capnp_struct as schemas
import os
import capnp

transaction_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/transaction.capnp')
subblock_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/subblock.capnp')
blockdata_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/blockdata.capnp')

PENDING_NONCE_KEY = '__pn'
NONCE_KEY = '__n'


class NonceManager:
    def __init__(self, driver=ContractDriver()):
        self.driver = driver

    @staticmethod
    def update_nonce_hash(nonce_hash: dict, tx_payload: transaction_capnp.TransactionPayload):
        # Modifies the provided dict
        k = (tx_payload.processor, tx_payload.sender)
        current_nonce = nonce_hash.get(k)

        if current_nonce is None or current_nonce == tx_payload.nonce:
            nonce_hash[k] = tx_payload.nonce + 1

    @staticmethod
    def n_key(key, processor, sender):
        return ':'.join([key, processor.hex(), sender.hex()])

    # Nonce methods
    def get_pending_nonce(self, processor: bytes, sender: bytes):
        return self.driver.get(self.n_key(PENDING_NONCE_KEY, processor, sender))

    def get_nonce(self, processor: bytes, sender: bytes):
        return self.driver.get(self.n_key(NONCE_KEY, processor, sender))

    def set_pending_nonce(self, processor: bytes, sender: bytes, nonce: int):
        self.driver.set(self.n_key(PENDING_NONCE_KEY, processor, sender), nonce)

    def set_nonce(self, processor: bytes, sender: bytes, nonce: int):
        self.driver.set(self.n_key(NONCE_KEY, processor, sender), nonce)

    def delete_pending_nonce(self, processor: bytes, sender: bytes):
        self.driver.delete(self.n_key(PENDING_NONCE_KEY, processor, sender))

    def commit_nonces(self, nonce_hash=None):
        # Delete pending nonces and update the nonces
        if nonce_hash is not None:
            for k, v in nonce_hash.items():
                processor, sender = k

                self.set_nonce(processor=processor, sender=sender, nonce=v)
                self.delete_pending_nonce(processor=processor, sender=sender)
        else:
            # Commit all pending nonces straight up
            for n in self.driver.iter(PENDING_NONCE_KEY):
                _, processor, sender = n.split(':')

                processor = bytes.fromhex(processor)
                sender = bytes.fromhex(sender)

                nonce = self.get_pending_nonce(processor=processor, sender=sender)

                self.set_nonce(processor=processor, sender=sender, nonce=nonce)
                self.driver.delete(n)

        self.driver.commit()

    def delete_pending_nonces(self):
        for nonce in self.driver.iter(PENDING_NONCE_KEY):
            self.driver.delete(nonce)

        self.driver.commit()

    def update_nonces_with_block(self, block):
        # Reinitialize the latest nonce. This should probably be abstracted into a seperate class at a later date
        nonces = {}

        for sb in block.subBlocks:
            for tx in sb.transactions:
                self.update_nonce_hash(nonce_hash=nonces, tx_payload=tx.transaction.payload)

        self.commit_nonces(nonce_hash=nonces)
        self.delete_pending_nonces()



def is_valid_hex(hex_str: str, length=0) -> bool:
    """
    Returns true if hex_str is valid hex. False otherwise
    :param hex_str: The string to check
    :param length: If set, also verify that hex_str is the valid length
    :return: A bool, true if hex_str is valid hex
    """
    try:
        assert isinstance(hex_str, str)
        int(hex_str, 16)
        if length:
            assert len(hex_str) == length
        return True
    except:
        return False


def int_to_bytes(x):
    return x.to_bytes((x.bit_length() + 7) // 8, 'big')




from unittest import TestCase
from cilantro_ee.core.sockets.services import _socket
from cilantro_ee.services.overlay.discovery import *
from cilantro_ee.services.overlay.network import Network, PeerServer, KTable
from cilantro_ee.services.overlay.discovery import DiscoveryServer
from cilantro_ee.constants.overlay_network import PEPPER
from cilantro_ee.core.sockets import services
import zmq
import zmq.asyncio
from cilantro_ee.core.crypto.wallet import Wallet
from time import sleep
import json

TIME_UNIT = 0.01


async def stop_server(s, timeout):
    await asyncio.sleep(timeout)
    s.stop()


def run_silent_loop(tasks, s=TIME_UNIT):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(tasks)
    except RuntimeError as e:
        pass
    sleep(s)


async def timeout_bomb(s=TIME_UNIT*2):
    await asyncio.sleep(s)
    asyncio.get_event_loop().close()


class TestNetworkService(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.ctx.destroy()
        self.loop.close()

    def test_setup(self):
        w = Wallet()
        Network(wallet=w, ctx=self.ctx)

    def test_bootstrap_nodes(self):
        w = Wallet()

        w1 = Wallet()
        w2 = Wallet()
        w3 = Wallet()
        w4 = Wallet()

        bootnodes = [_socket('tcp://127.0.0.1:10999'),
                     _socket('tcp://127.0.0.1:11999'),
                     _socket('tcp://127.0.0.1:12999'),
                     _socket('tcp://127.0.0.1:13999')]

        d1 = DiscoveryServer(bootnodes[0], w1, pepper=PEPPER.encode(), ctx=self.ctx, linger=1000, poll_timeout=1000)
        d2 = DiscoveryServer(bootnodes[1], w2, pepper=PEPPER.encode(), ctx=self.ctx, linger=1000, poll_timeout=1000)
        d3 = DiscoveryServer(bootnodes[2], w3, pepper=PEPPER.encode(), ctx=self.ctx, linger=1000, poll_timeout=1000)
        d4 = DiscoveryServer(bootnodes[3], w4, pepper=PEPPER.encode(), ctx=self.ctx, linger=1000, poll_timeout=1000)

        n = Network(wallet=w, ctx=self.ctx, bootnodes=bootnodes)

        tasks = asyncio.gather(
            d1.serve(),
            d2.serve(),
            d3.serve(),
            d4.serve(),
            stop_server(d1, 1),
            stop_server(d2, 1),
            stop_server(d3, 1),
            stop_server(d4, 1),
            n.discover_bootnodes(bootnodes)
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

        expected_dict = {
            w1.verifying_key().hex(): 'tcp://127.0.0.1:10999',
            w2.verifying_key().hex(): 'tcp://127.0.0.1:11999',
            w3.verifying_key().hex(): 'tcp://127.0.0.1:12999',
            w4.verifying_key().hex(): 'tcp://127.0.0.1:13999'
        }

        self.assertDictEqual(n.table.peers, expected_dict)

    def test_peer_server_init(self):
        w = Wallet()
        t = KTable(data={'woo': 'hoo'})
        p = PeerServer(socket_id=_socket('tcp://127.0.0.1:10999'), wallet=w, event_port=8888,
                       table=t, ctx=self.ctx, linger=100, poll_timeout=100, discovery_port=9999)

        tasks = asyncio.gather(
            p.serve(),
            stop_server(p, 0.1)
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

    def test_peer_server_returns_self_when_asked(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ip='127.0.0.1', ctx=self.ctx, peer_service_port=10001, event_publisher_port=10002)

        find_message = ['find', w1.verifying_key().hex()]
        find_message = json.dumps(find_message).encode()

        tasks = asyncio.gather(
            p1.peer_service.serve(),
            stop_server(p1.peer_service, 0.3),
            services.get(_socket('tcp://127.0.0.1:10001'), msg=find_message, ctx=self.ctx, timeout=300)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        response = res[-1]
        response = response.decode()
        response = json.loads(response)

        self.assertEqual(response.get(w1.verifying_key().hex()), '127.0.0.1')

    def test_peer_server_returns_peer_when_asked(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ip='127.0.0.1', ctx=self.ctx, peer_service_port=10001, event_publisher_port=10002)

        w2 = Wallet()

        p1.peer_service.table.peers[w2.verifying_key().hex()] = 'inproc://goodtimes'

        find_message = ['find', w2.verifying_key().hex()]
        find_message = json.dumps(find_message).encode()

        tasks = asyncio.gather(
            p1.peer_service.serve(),
            stop_server(p1.peer_service, 0.3),
            services.get(_socket('tcp://127.0.0.1:10001'), msg=find_message, ctx=self.ctx, timeout=300)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        response = res[-1]
        response = response.decode()
        response = json.loads(response)

        self.assertEqual(response.get(w2.verifying_key().hex()), 'inproc://goodtimes')

    def test_peer_server_returns_all_peers_if_doesnt_have_it_or_more_than_response_amount(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ip='127.0.0.1', ctx=self.ctx, peer_service_port=10001, event_publisher_port=10002)

        test_dict = {
            'test': 'value',
            'another': 'one',
            'something': 'else'
        }

        p1.peer_service.table.peers = test_dict

        find_message = ['find', 'baloney']
        find_message = json.dumps(find_message).encode()

        tasks = asyncio.gather(
            p1.peer_service.serve(),
            stop_server(p1.peer_service, 0.3),
            services.get(_socket('tcp://127.0.0.1:10001'), msg=find_message, ctx=self.ctx, timeout=300)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        response = res[-1]
        response = response.decode()
        response = json.loads(response)

        self.assertDictEqual(test_dict, response)

    def test_peer_server_returns_max_response_keys(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ip='127.0.0.1', ctx=self.ctx, peer_service_port=10001, event_publisher_port=10002)

        test_dict = {
            'test': 'value',
            'another': 'one',
            'something': 'else'
        }

        expected_dict = {
            'test': 'value'
        }

        p1.peer_service.table.peers = test_dict
        p1.peer_service.table.response_size = 1

        find_message = ['find', 'tesZ']
        find_message = json.dumps(find_message).encode()

        tasks = asyncio.gather(
            p1.peer_service.serve(),
            stop_server(p1.peer_service, 0.2),
            services.get(_socket('tcp://127.0.0.1:10001'), msg=find_message, ctx=self.ctx, timeout=200)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        response = res[-1]
        response = response.decode()
        response = json.loads(response)

        self.assertDictEqual(expected_dict, response)

    def test_peer_server_returns_max_response_keys_many_keys(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ip='127.0.0.1', ctx=self.ctx, peer_service_port=10001, event_publisher_port=10002)

        test_dict = {
            '0': 'value',
            '1': 'one',
            '2': 'else',
            '3': 'value',
            '4': 'one',
            '5': 'else',
            '7': 'one',
            '8': 'else',
            '9': 'value',
            '10': 'one',
            '11': 'else',
            '12': 'value',
            '13': 'one',
            '14': 'else',
        }

        expected_dict = {
            '4': 'one',
            '5': 'else',
            '7': 'one',
            '2': 'else'
        }

        p1.peer_service.table.peers = test_dict
        p1.peer_service.table.response_size = 4

        find_message = ['find', '6']
        find_message = json.dumps(find_message).encode()

        tasks = asyncio.gather(
            p1.peer_service.serve(),
            stop_server(p1.peer_service, 0.2),
            services.get(_socket('tcp://127.0.0.1:10001'), msg=find_message, ctx=self.ctx, timeout=200)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        response = res[-1]
        response = response.decode()
        response = json.loads(response)

        self.assertDictEqual(expected_dict, response)

    def test_peer_table_updated_on_join_command(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ip='127.0.0.1', ctx=self.ctx, peer_service_port=10001, event_publisher_port=10002)

        w2 = Wallet()
        d = DiscoveryServer(wallet=w2, socket_id=_socket('tcp://127.0.0.1:10999'), pepper=PEPPER.encode(), ctx=self.ctx, linger=200)

        # 1. start network
        # 2. start discovery of other side
        # 3. send join request
        # 4. check to see if the data has been added

        join_message = ['join', (w2.verifying_key().hex(), 'tcp://127.0.0.1:10999')]
        join_message = json.dumps(join_message).encode()

        tasks = asyncio.gather(
            p1.peer_service.serve(),
            d.serve(),
            services.get(_socket('tcp://127.0.0.1:10001'), msg=join_message, ctx=self.ctx, timeout=1000),
            stop_server(p1.peer_service, 0.3),
            stop_server(d, 0.3)
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

        self.assertEqual(p1.peer_service.table.peers[w2.verifying_key().hex()], 'tcp://127.0.0.1:10999')

    def test_event_service_publisher_starts_up_on_init(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        test_subscriber = self.ctx.socket(zmq.SUB)
        test_subscriber.setsockopt(zmq.SUBSCRIBE, b'')
        test_subscriber.connect('tcp://127.0.0.1:10002')

        # TCP takes a bit longer to bind and is prone to dropping messages...
        sleep(0.1)

        async def send():
            await p1.peer_service.event_publisher.send(b'waaaa')

        async def recv():
            return await test_subscriber.recv()

        tasks = asyncio.gather(
            send(),
            recv(),
            stop_server(p1.peer_service, 0.1)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        self.assertEqual(res[1], b'waaaa')

    def test_event_service_triggered_when_new_node_added(self):
        # Create Network service
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        # Create Discovery Server
        w2 = Wallet()
        d = DiscoveryServer(wallet=w2, socket_id=_socket('tcp://127.0.0.1:10999'), pepper=PEPPER.encode(), ctx=self.ctx,
                            poll_timeout=2000, linger=200)

        # Create raw subscriber
        subscriber = self.ctx.socket(zmq.SUB)
        subscriber.setsockopt(zmq.SUBSCRIBE, b'')
        subscriber.connect('tcp://127.0.0.1:10002')

        # TCP takes a bit longer to bind and is prone to dropping messages...
        sleep(0.3)

        # Construct the join RPC message
        join_message = ['join', (w2.verifying_key().hex(), 'tcp://127.0.0.1:10999')]
        join_message = json.dumps(join_message).encode()

        # Wrap recv() in an async
        async def recv():
            msg = await subscriber.recv()
            return msg

        tasks = asyncio.gather(
            p1.peer_service.start(),  # Start the PeerService which will process RPC and emit events
            d.serve(),  # Start Discovery so PeerService can verify they are online
            services.get(_socket('tcp://127.0.0.1:10001'), msg=join_message, ctx=self.ctx, timeout=3000),  # Push out a join request
            stop_server(p1.peer_service, 1),
            stop_server(d, 1),
            recv()  # Collect the subscription result
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        expected_list = ['join', [w2.verifying_key().hex(), 'tcp://127.0.0.1:10999']]
        got_list = json.loads(res[-1].decode())

        self.assertListEqual(expected_list, got_list)

    def test_other_peers_add_new_nodes_when_join_event_occurs(self):
        # Create Network service
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002, discovery_port=10999)

        # Create Network service
        w2 = Wallet()
        p2 = Network(wallet=w2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        p2.peer_service.event_service.add_subscription(_socket('tcp://127.0.0.1:10002'))

        # Create Discovery Server
        w3 = Wallet()
        d = DiscoveryServer(wallet=w3, socket_id=_socket('tcp://127.0.0.1:10999'), pepper=PEPPER.encode(), ctx=self.ctx,
                            poll_timeout=2000, linger=2000)

        # TCP takes a bit longer to bind and is prone to dropping messages...
        sleep(1)

        # Construct the join RPC message
        join_message = ['join', (w3.verifying_key().hex(), 'tcp://127.0.0.1:10999')]
        join_message = json.dumps(join_message).encode()

        tasks = asyncio.gather(
            p1.peer_service.start(),
            p2.peer_service.start(),
            d.serve(),
            services.get(_socket('tcp://127.0.0.1:10001'), msg=join_message, ctx=self.ctx, timeout=1000),
            stop_server(p1.peer_service, 2),
            stop_server(p2.peer_service, 2),
            stop_server(d, 2),
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

        self.assertTrue(w3.verifying_key().hex() in p2.peer_service.table.peers)

    def test_start_and_stopping_destroys_servers(self):
        # Create Network service
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        # Create Network service
        w2 = Wallet()
        p2 = Network(wallet=w2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        async def stop(n: Network, s):
            await asyncio.sleep(s)
            n.peer_service.stop()

        tasks = asyncio.gather(
            p1.peer_service.start(),

            p2.peer_service.start(),

            stop(p1, 0.3),
            stop(p2, 0.3)
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

    def test_find_node_returns_self_if_asked_on_peer_address_and_self_is_the_value(self):
        # Create Network service
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        async def get():
            return await p1.find_node(p1.peer_service_address, w1.verifying_key().hex())

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(get())

        self.assertEqual(res.get(w1.verifying_key().hex()), '127.0.0.1')

    def test_find_node_gets_node_from_self_if_asked_from_self_and_has_it_as_peer(self):
        # Create Network service
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        w2 = Wallet()
        p1.table.peers[w2.verifying_key().hex()] = '9.9.9.9'

        async def get():
            return await p1.find_node(p1.peer_service_address, w2.verifying_key().hex())

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(get())

        self.assertEqual(res.get(w2.verifying_key().hex()), '9.9.9.9')

    def test_find_node_requests_from_others_and_returns_key_if_they_have_it(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        w2 = Wallet()
        p2 = Network(wallet=w2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        async def get():
            return await p1.find_node(_socket('tcp://127.0.0.1:10003'), w2.verifying_key().hex())

        async def stop(n: Network, s):
            await asyncio.sleep(s)
            n.peer_service.stop()

        tasks = asyncio.gather(
            p1.peer_service.start(),

            p2.peer_service.start(),
            get(),
            stop(p1, 0.3),
            stop(p2, 0.3),

        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        self.assertEqual(res[2].get(w2.verifying_key().hex()), '127.0.0.1')

    def test_find_node_fails_if_cant_find_and_retries_are_up(self):
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        w2 = Wallet()
        p2 = Network(wallet=w2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        w3 = Wallet()

        async def get():
            return await p1.find_node(_socket('tcp://127.0.0.1:10003'), w3.verifying_key().hex(), retries=1)

        async def stop(n: Network, s):
            await asyncio.sleep(s)
            n.peer_service.stop()

        tasks = asyncio.gather(
            p1.peer_service.start(),

            p2.peer_service.start(),
            get(),
            stop(p1, 0.3),
            stop(p2, 0.3),

        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        self.assertIsNone(res[2])

    def test_recursive_crawl_works_to_proper_depth(self):
        # 0 tries
        w1 = Wallet()
        p1 = Network(wallet=w1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        # 1 try
        w2 = Wallet()
        p2 = Network(wallet=w2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        w3 = Wallet()
        p3 = Network(wallet=w3, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10005, event_publisher_port=10006)

        # 2 tries <- info in this node should be returned
        w4 = Wallet()
        p4 = Network(wallet=w4, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10007, event_publisher_port=10008)

        # 2 tries <- info in this node should be returned
        w5 = Wallet()

        # Add node info in each peer service to 'chain' them together
        p2.table.peers[w3.verifying_key().hex()] = 'tcp://127.0.0.1:10005'
        p3.table.peers[w4.verifying_key().hex()] = 'tcp://127.0.0.1:10007'
        p4.table.peers[w5.verifying_key().hex()] = 'you found me!'

        async def get():
            return await p1.find_node(_socket('tcp://127.0.0.1:10003'), w5.verifying_key().hex(), retries=2)

        async def stop(n: Network, s):
            await asyncio.sleep(s)
            n.peer_service.stop()

        timeout = 0.3

        tasks = asyncio.gather(
            p1.peer_service.start(),
            p2.peer_service.start(),
            p3.peer_service.start(),
            p4.peer_service.start(),
            get(),
            stop(p1, timeout),
            stop(p2, timeout),
            stop(p3, timeout),
            stop(p4, timeout),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        self.assertEqual(res[4].get(w5.verifying_key().hex()), 'you found me!')

    def test_wait_for_quorum_to_succeed_only_one_master(self):
        # 0 tries
        mnw1 = Wallet()
        mn1 = Network(wallet=mnw1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        # 1 try
        mnw2 = Wallet()
        mn2 = Network(wallet=mnw2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        dw1 = Wallet()
        d1 = Network(wallet=dw1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10005, event_publisher_port=10006)

        # 2 tries <- info in this node should be returned
        dw2 = Wallet()
        d2 = Network(wallet=dw2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10007, event_publisher_port=10008)

        async def get():
            return await mn1.wait_for_quorum(1, 0, [mnw1.verifying_key().hex(), mnw2.verifying_key().hex()],
                                             [dw1.verifying_key().hex(), dw2.verifying_key().hex()],
                                             initial_peers=[_socket('tcp://127.0.0.1:10003')])

        async def stop(n: Network, s):
            await asyncio.sleep(s)
            n.peer_service.stop()

        timeout = 0.3

        tasks = asyncio.gather(
            mn1.peer_service.start(),
            mn2.peer_service.start(),
            d1.peer_service.start(),
            d2.peer_service.start(),
            get(),
            stop(mn1, timeout),
            stop(mn2, timeout),
            stop(d1, timeout),
            stop(d1, timeout),
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

        self.assertIn(mnw2.verifying_key().hex(), mn1.table.peers)

    def test_wait_for_quorum_with_initial_nodes_being_single_node(self):
        # Setup nodes
        mnw1 = Wallet()
        mn1 = Network(wallet=mnw1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        mnw2 = Wallet()
        mn2 = Network(wallet=mnw2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        dw1 = Wallet()
        d1 = Network(wallet=dw1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10005, event_publisher_port=10006)

        dw2 = Wallet()
        d2 = Network(wallet=dw2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10007, event_publisher_port=10008)

        # Add various info to each so that a crawl is successful
        mn2.table.peers[dw1.verifying_key().hex()] = 'tcp://127.0.0.1:10005'
        d1.table.peers[dw2.verifying_key().hex()] = 'tcp://127.0.0.1:10007'

        async def get():
            return await mn1.wait_for_quorum(2, 2, [mnw1.verifying_key().hex(), mnw2.verifying_key().hex()],
                                             [dw1.verifying_key().hex(), dw2.verifying_key().hex()],
                                             initial_peers=[_socket('tcp://127.0.0.1:10003')])

        async def stop(n: Network, s):
            await asyncio.sleep(s)
            n.peer_service.stop()

        timeout = 0.3

        tasks = asyncio.gather(
            mn1.peer_service.start(),
            mn2.peer_service.start(),
            d1.peer_service.start(),
            d2.peer_service.start(),
            get(),
            stop(mn1, timeout),
            stop(mn2, timeout),
            stop(d1, timeout),
            stop(d1, timeout),
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

        self.assertIn(mnw2.verifying_key().hex(), mn1.table.peers)
        self.assertIn(dw1.verifying_key().hex(), mn1.table.peers)
        self.assertIn(dw2.verifying_key().hex(), mn1.table.peers)

    def test_wait_for_quorum_resolves_when_late_joiner(self):
        # Setup nodes
        mnw1 = Wallet()
        mn1 = Network(wallet=mnw1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10001, event_publisher_port=10002)

        mnw2 = Wallet()
        mn2 = Network(wallet=mnw2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10003, event_publisher_port=10004)

        dw1 = Wallet()
        d1 = Network(wallet=dw1, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10005, event_publisher_port=10006)

        dw2 = Wallet()
        d2 = Network(wallet=dw2, ctx=self.ctx, ip='127.0.0.1', peer_service_port=10007, event_publisher_port=10008)

        # Add various info to each so that a crawl is successful
        mn2.table.peers[dw1.verifying_key().hex()] = 'tcp://127.0.0.1:10005'
        d1.table.peers[dw2.verifying_key().hex()] = 'tcp://127.0.0.1:10007'

        async def get():
            return await mn1.wait_for_quorum(2, 2, [mnw1.verifying_key().hex(), mnw2.verifying_key().hex()],
                                             [dw1.verifying_key().hex(), dw2.verifying_key().hex()],
                                             initial_peers=[_socket('tcp://127.0.0.1:10003')])

        async def stop(n: Network, s):
            await asyncio.sleep(s)
            n.peer_service.stop()

        async def start_late(n: Network, s):
            await asyncio.sleep(s)
            await n.peer_service.start()

        timeout = 0.5

        tasks = asyncio.gather(
            mn1.peer_service.start(),
            start_late(mn2, 0.1),
            start_late(d1, 0.2),
            start_late(d2, 0.3),
            get(),
            stop(mn1, timeout),
            stop(mn2, timeout),
            stop(d1, timeout),
            stop(d1, timeout),
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

        self.assertIn(mnw2.verifying_key().hex(), mn1.table.peers)
        self.assertIn(dw1.verifying_key().hex(), mn1.table.peers)
        self.assertIn(dw2.verifying_key().hex(), mn1.table.peers)

    # def test quorum made and another node wants to connect afterwards


from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('2-2-2.json')
from cilantro_ee.constants.testnet import set_testnet_nodes
set_testnet_nodes()

from cilantro_ee.nodes.masternode.block_aggregator import BlockAggregator

import unittest, asyncio
from unittest import TestCase, mock
from unittest.mock import MagicMock

from cilantro_ee.constants.ports import MN_PUB_PORT
from cilantro_ee.constants.system_config import *

from cilantro_ee.messages.envelope.envelope import Envelope
from cilantro_ee.messages.consensus.sub_block_contender import SubBlockContender
from cilantro_ee.messages.transaction.contract import ContractTransactionBuilder

from cilantro_ee.messages.block_data.block_data import *
from cilantro_ee.messages.block_data.state_update import *
from cilantro_ee.messages.block_data.block_metadata import *
from cilantro_ee.messages.consensus.merkle_signature import *

from cilantro_ee.core.containers.merkle_tree import MerkleTree
from cilantro_ee.services.storage.driver import SafeDriver
from cilantro_ee.services.storage.vkbook import PhoneBook, VKBook

# time and logger are for debugging
from cilantro_ee.core.logger.base import get_logger


class BlockAggTester:
    @staticmethod
    def test(func):
        @mock.patch("cilantro_ee.nodes.masternode.block_aggregator.NUM_SB_PER_BLOCK", 2)
        @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 2)
        @mock.patch("cilantro_ee.nodes.masternode.block_contender.NUM_SB_PER_BLOCK", 2)
        # @mock.patch("cilantro_ee.core.utils.worker.asyncio")
        @mock.patch("cilantro_ee.core.utils.worker.SocketManager")
        @mock.patch("cilantro_ee.nodes.masternode.block_aggregator.BlockAggregator.run")
        def _func(*args, **kwargs):
            return func(*args, **kwargs)
        return _func


TEST_IP = '127.0.0.1'
TEST_SK = TESTNET_MASTERNODES[0]['sk']
TEST_VK = TESTNET_MASTERNODES[0]['vk']
DEL_SK = TESTNET_DELEGATES[0]['sk']
DEL_VK = TESTNET_DELEGATES[0]['vk']

TEST_DELEGATES = [TESTNET_DELEGATES[i]['vk'] for i in range(len(TESTNET_DELEGATES))]
TEST_MASTERS = [TESTNET_MASTERNODES[i]['vk'] for i in range(len(TESTNET_MASTERNODES))]

INPUT_HASH1 = '1' * 64
INPUT_HASH2 = '2' * 64
RAWTXS1 = [
    b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA',
    b'BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB',
    b'CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC',
    b'DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD',
    b'EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE'
]
RAWTXS2 = [
    b'1AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA',
    b'2BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB',
    b'3CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC',
    b'4DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD',
    b'5EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE'
]
TXS1 = [TransactionData.create(
    contract_tx=ContractTransactionBuilder.create_currency_tx(sender_sk=TEST_SK, receiver_vk=raw_tx.decode(), amount=10),
    status='SUCCESS', state='x 1') for raw_tx in RAWTXS1]
TXS2 = [TransactionData.create(
    contract_tx=ContractTransactionBuilder.create_currency_tx(sender_sk=TEST_SK, receiver_vk=raw_tx.decode(), amount=10),
    status='SUCCESS', state='x 1') for raw_tx in RAWTXS2]

TREE1 = MerkleTree.from_raw_transactions([t.serialize() for t in TXS1])
TREE2 = MerkleTree.from_raw_transactions([t.serialize() for t in TXS2])

MERKLE_LEAVES1 = TREE1.leaves
MERKLE_LEAVES2 = TREE2.leaves

RESULT_HASH1 = TREE1.root_as_hex
RESULT_HASH2 = TREE2.root_as_hex

log = get_logger('BlockAggregatorTester')


class TestBlockAggregator(TestCase):
    def setUp(self):
        SafeDriver.flush()

    @BlockAggTester.test
    def test_build_task_list_connect_and_bind(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        # tasks is usually self.manager.overlay_client.tasks, (which in this case is mocked out), so we must do thus:
        ba.tasks = []

        mock_manager = MagicMock()
        ba.manager = mock_manager

        mock_pub, mock_sub, mock_router, mock_ipc_router = MagicMock(), MagicMock(), MagicMock(), MagicMock()
        mock_manager.create_socket = MagicMock(side_effect=[mock_sub, mock_pub, mock_router, mock_ipc_router])
        mock_manager.is_ready = MagicMock(return_value=True)

        mock_sub_handler_task = MagicMock()
        mock_sub.add_handler = MagicMock(return_value=mock_sub_handler_task)

        ba.build_task_list()
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        self.assertEqual(ba.sub, mock_sub)
        self.assertEqual(ba.pub, mock_pub)
        mock_sub.add_handler.assert_called()
        self.assertTrue(mock_sub_handler_task in ba.tasks)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(ba._connect_and_process())
        loop.close()

        expected_num_mn_subs = len([vk for vk in PhoneBook.masternodes if TEST_VK != vk])
        expected_num_delegate_subs = len(PhoneBook.delegates)
        self.assertEqual(mock_sub.connect.call_count, expected_num_mn_subs + expected_num_delegate_subs)

        mock_pub.bind.assert_called_with(ip=TEST_IP, port=MN_PUB_PORT)

    # TODO fix this test --davis
    # @BlockAggTester.test
    # def test_store_block(self, *args):
    #     ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)
    #     ba.manager = MagicMock()
    #     ba.send_new_block_notif = MagicMock()
    #     ba.build_task_list()
    #     ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)
    #
    #     sk1, vk1 = wallet._new()
    #     sk2, vk2 = wallet._new()
    #     sk3, vk3 = wallet._new()
    #     sk4, vk4 = wallet._new()
    #
    #     sb1_txs = [TransactionDataBuilder.create_random_tx() for _ in range(8)]
    #     sb2_txs = [TransactionDataBuilder.create_random_tx() for _ in range(8)]
    #     sb1_txs_hashes = {Hasher.hash(tx): tx for tx in sb1_txs}
    #     sb2_txs_hashes = {Hasher.hash(tx): tx for tx in sb2_txs}
    #
    #     tree1 = MerkleTree.from_transactions(sb1_txs)
    #     tree2 = MerkleTree.from_transactions(sb2_txs)
    #
    #     sig1, sig2 = MerkleSignature.create_from_payload(sk1, tree1.root), MerkleSignature.create_from_payload(sk2, tree1.root)
    #     sig3, sig4 = MerkleSignature.create_from_payload(sk3, tree2.root), MerkleSignature.create_from_payload(sk4, tree2.root)
    #
    #     input_hash1 = 'AB' * 32  # Input hashes are the hashes of the transaction bags
    #     input_hash2 = 'BA' * 32
    #     result_hash1 = tree1.root_as_hex  # Result hashes are the merkle roots
    #     result_hash2 = tree2.root_as_hex
    #
    #     ba.result_hashes[result_hash1] = {'_committed_': False, '_consensus_reached_': True,
    #                                       '_transactions_': sb1_txs_hashes,
    #                                       '_valid_signatures_': {sig1.signature: sig1, sig2.signature: sig2},
    #                                       '_input_hash_': input_hash1, '_merkle_leaves_': tree1.leaves_as_hex,
    #                                       '_sb_index_': 0, '_lastest_valid_': time.time()}
    #     ba.result_hashes[result_hash2] = {'_committed_': False, '_consensus_reached_': True,
    #                                       '_transactions_': sb2_txs_hashes,
    #                                       '_valid_signatures_': {sig3.signature: sig3, sig4.signature: sig4},
    #                                       '_input_hash_': input_hash2, '_merkle_leaves_': tree2.leaves_as_hex,
    #                                       '_sb_index_': 1, '_lastest_valid_': time.time()}
    #
    #     ba.store_full_block()
    #
    #     block_data = ba.send_new_block_notif.call_args[0][0]
    #     block_data = ba.send_new_block_notif.call_args[0][0]
    #
    #     self.assertEqual(block_data.sub_blocks[0].input_hash, input_hash1)
    #     self.assertEqual(block_data.sub_blocks[1].input_hash, input_hash2)
    #     self.assertEqual(block_data.sub_blocks[0].merkle_root, result_hash1)
    #     self.assertEqual(block_data.sub_blocks[1].merkle_root, result_hash2)
    #     self.assertEqual(block_data.sub_blocks[0].transactions, sb1_txs)
    #     self.assertEqual(block_data.sub_blocks[1].transactions, sb2_txs)
    #     self.assertEqual(block_data.sub_blocks[0].signatures, [sig1, sig2])
    #     self.assertEqual(block_data.sub_blocks[1].signatures, [sig3, sig4])

    @BlockAggTester.test
    def test_handle_sub_msg_with_sub_block_contender(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.recv_sub_block_contender = MagicMock()
        ba.build_task_list()
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        mock_env = MagicMock()
        mock_env.message = MagicMock(spec=SubBlockContender)

        with mock.patch.object(Envelope, 'from_bytes', return_value=mock_env):
            ba.handle_sub_msg([b'filter doesnt matter', b'envelope binary also doesnt matter'])

        ba.recv_sub_block_contender.assert_called()

    @BlockAggTester.test
    def test_handle_sub_msg_with_new_block_notif(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.recv_new_block_notif = MagicMock()
        ba.build_task_list()
    
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        mock_env = MagicMock()
        mock_env.message = MagicMock(spec=NewBlockNotification)

        with mock.patch.object(Envelope, 'from_bytes', return_value=mock_env):
            ba.handle_sub_msg([b'filter doesnt matter', b'envelope binary also doesnt matter'])

        ba.recv_new_block_notif.assert_called()

    @BlockAggTester.test
    def test_recv_sub_block_contender(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.build_task_list()
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH1), sk=DEL_SK, vk=DEL_VK)

        PhoneBook = VKBook(delegates=[DEL_VK, TESTNET_DELEGATES[1]['vk']], masternodes=[TESTNET_MASTERNODES[0]])

        sbc = SubBlockContender.create(RESULT_HASH1, INPUT_HASH1, MERKLE_LEAVES1, signature, TXS1, 0, GENESIS_BLOCK_HASH)

        ba.recv_sub_block_contender(DEL_VK, sbc)
        self.assertTrue(sbc in ba.curr_block.sb_groups[0].rh[RESULT_HASH1])


    @BlockAggTester.test
    def test_recv_empty_sub_block_contender(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.build_task_list()
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        signature = build_test_merkle_sig(msg=bytes.fromhex(INPUT_HASH1), sk=DEL_SK, vk=DEL_VK)

        PhoneBook = VKBook(delegates=[DEL_VK, TESTNET_DELEGATES[1]['vk']], masternodes=[TESTNET_MASTERNODES[0]])

        sbc = SubBlockContender.create_empty_sublock(INPUT_HASH1, sub_block_index=0, signature=signature,
                                                     prev_block_hash=GENESIS_BLOCK_HASH)

    @BlockAggTester.test
    @mock.patch("cilantro_ee.nodes.masternode.block_aggregator.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.nodes.masternode.block_contender.NUM_SB_PER_BLOCK", 1)
    def test_combine_result_hash_with_one_sb(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.send_new_block_notif = MagicMock()
        ba.build_task_list()
        ba.pub = MagicMock()
        old_b_hash = ba.curr_block_hash
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        PhoneBook = VKBook(delegates=TEST_DELEGATES, masternodes=TEST_MASTERS)

        # Sub block 0
        for i in range(DELEGATE_MAJORITY):
            signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH1), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
            sbc = SubBlockContender.create(RESULT_HASH1, INPUT_HASH1, MERKLE_LEAVES1, signature, TXS1, 0, GENESIS_BLOCK_HASH)
            ba.recv_sub_block_contender(TESTNET_DELEGATES[i]['vk'], sbc)

        block_data = ba.send_new_block_notif.call_args[0][0]
        self.assertEqual(block_data.prev_block_hash, old_b_hash)
        self.assertEqual(block_data.merkle_roots, [RESULT_HASH1])
        self.assertEqual(block_data.transactions, TXS1)

    @BlockAggTester.test
    def test_combine_result_hash_with_multiple_sb(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.build_task_list()
        ba.send_new_block_notif = MagicMock()
        old_b_hash = ba.curr_block_hash

        PhoneBook = VKBook(delegates=TEST_DELEGATES, masternodes=TEST_MASTERS)

        # Sub block 0
        for i in range(DELEGATE_MAJORITY):
            signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH1), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
            sbc = SubBlockContender.create(RESULT_HASH1, INPUT_HASH1, MERKLE_LEAVES1, signature, TXS1, 0, GENESIS_BLOCK_HASH)
            ba.recv_sub_block_contender(TESTNET_DELEGATES[i]['vk'], sbc)

        # Sub block 1
        for i in range(DELEGATE_MAJORITY):
            signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH2), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
            sbc = SubBlockContender.create(RESULT_HASH2, INPUT_HASH2, MERKLE_LEAVES2, signature, TXS2, 1, GENESIS_BLOCK_HASH)
            ba.recv_sub_block_contender(TESTNET_DELEGATES[i]['vk'], sbc)

        block_data = ba.send_new_block_notif.call_args[0][0]
        self.assertEqual(block_data.prev_block_hash, old_b_hash)
        self.assertEqual(block_data.merkle_roots, [RESULT_HASH1, RESULT_HASH2])
        self.assertEqual(block_data.transactions, TXS1 + TXS2)

    @BlockAggTester.test
    def test_combine_result_hash_with_multiple_sb_with_extras(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.send_new_block_notif = MagicMock()
        ba.build_task_list()
        ba.pub = MagicMock()
        old_b_hash = ba.curr_block_hash
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        PhoneBook = VKBook(delegates=TEST_DELEGATES, masternodes=TEST_MASTERS)

        # Sub block 0
        for i in range(NUM_DELEGATES):
            signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH1), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
            sbc = SubBlockContender.create(RESULT_HASH1, INPUT_HASH1, MERKLE_LEAVES1, signature, TXS1, 0, GENESIS_BLOCK_HASH)
            ba.recv_sub_block_contender(TESTNET_DELEGATES[i]['vk'], sbc)

        # Sub block 1
        for i in range(NUM_DELEGATES):
            signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH2), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
            sbc = SubBlockContender.create(RESULT_HASH2, INPUT_HASH2, MERKLE_LEAVES2, signature, TXS2, 1, GENESIS_BLOCK_HASH)
            ba.recv_sub_block_contender(TESTNET_DELEGATES[i]['vk'], sbc)

        block_data = ba.send_new_block_notif.call_args[0][0]
        self.assertEqual(block_data.prev_block_hash, old_b_hash)
        self.assertEqual(block_data.merkle_roots, [RESULT_HASH1, RESULT_HASH2])
        self.assertEqual(block_data.transactions, TXS1 + TXS2)

    @BlockAggTester.test
    @mock.patch("cilantro_ee.nodes.masternode.block_aggregator.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.nodes.masternode.block_contender.NUM_SB_PER_BLOCK", 1)
    def test_recv_ignore_extra_sub_block_contenders(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)
        ba.manager = MagicMock()
        ba.build_task_list()
        ba.send_new_block_notif = MagicMock()
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)
        old_b_hash = ba.curr_block_hash

        for i in range(DELEGATE_MAJORITY + 5):
            vk = TESTNET_DELEGATES[i%len(TESTNET_DELEGATES)]['vk']
            sk = TESTNET_DELEGATES[i%len(TESTNET_DELEGATES)]['sk']
            signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH1), sk=sk, vk=vk)
            sbc = SubBlockContender.create(RESULT_HASH1, INPUT_HASH1, MERKLE_LEAVES1, signature, TXS1, 0, GENESIS_BLOCK_HASH)
            ba.recv_sub_block_contender(vk, sbc)

        block_data = ba.send_new_block_notif.call_args[0][0]
        self.assertEqual(block_data.prev_block_hash, old_b_hash)
        self.assertEqual(block_data.merkle_roots, [RESULT_HASH1])
        self.assertEqual(block_data.transactions, TXS1)

    # TODO fix this test once we care about getting consensus on _new block notifications --davis
    # @BlockAggTester.test
    # @mock.patch("cilantro_ee.nodes.masternode.block_aggregator.MASTERNODE_MAJORITY", 2)
    # @mock.patch("cilantro_ee.nodes.masternode.block_aggregator.NUM_SB_PER_BLOCK", 2)
    # @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 2)
    # def test_recv_result_hash_multiple_subblocks_consensus(self, *args):
    #
    #     ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)
    #     ba.manager = MagicMock()
    #     ba.build_task_list()
    #     ba.is_catching_up = False
    #     ba.pub = MagicMock()
    #     bh = ba.curr_block_hash
    #
    #     # Sub block 0
    #     for i in range(DELEGATE_MAJORITY):
    #         signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH1), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
    #         sbc = SubBlockContender.create(RESULT_HASH1, INPUT_HASH1, MERKLE_LEAVES1, signature, TXS1, 0, GENESIS_BLOCK_HASH)
    #         ba.recv_sub_block_contender(sbc)
    #
    #     # Sub block 1
    #     for i in range(DELEGATE_MAJORITY):
    #         signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH2), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
    #         sbc = SubBlockContender.create(RESULT_HASH2, INPUT_HASH2, MERKLE_LEAVES2, signature, TXS2, 1, GENESIS_BLOCK_HASH)
    #         ba.recv_sub_block_contender(sbc)
    #
    #     self.assertEqual(len(ba.full_blocks), 1)
    #     block_data = ba.send_new_block_notif.call_args[0][0]
    #     nbn = NewBlockNotification.create_from_block_data(block_data)
    #
    #     mn1_vk = 'ABCD' * 16
    #     mn2_vk = 'DCBA' * 16
    #     ba.recv_new_block_notif(mn1_vk)
    #     ba.recv_new_block_notif(mn2_vk)
    #
    #     self.assertEqual(len(ba.full_blocks[ba.curr_block_hash]['_senders_']), 3)
    #     self.assertEqual(block_data.prev_block_hash, bh)
    #     self.assertEqual(block_data.merkle_roots, [RESULT_HASH1, RESULT_HASH2])
    #     self.assertEqual(block_data.transactions, TXS1 + TXS2)

    @BlockAggTester.test
    @mock.patch("cilantro_ee.nodes.masternode.block_aggregator.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.nodes.masternode.block_contender.NUM_SB_PER_BLOCK", 1)
    def test_combine_result_hash_transactions_missing(self, *args):
        ba = BlockAggregator(ip=TEST_IP, signing_key=TEST_SK, ipc_ip="test_mn-ipc-sock", ipc_port=6967)

        ba.manager = MagicMock()
        ba.build_task_list()
        ba.send_new_block_notif = MagicMock()
        ba.catchup_manager.is_catchup_done = MagicMock(return_value=True)

        PhoneBook = VKBook(delegates=TEST_DELEGATES, masternodes=TEST_MASTERS)

        for i in range(DELEGATE_MAJORITY):
            signature = build_test_merkle_sig(msg=bytes.fromhex(RESULT_HASH1), sk=TESTNET_DELEGATES[i]['sk'], vk=TESTNET_DELEGATES[i]['vk'])
            sbc = SubBlockContender.create(RESULT_HASH1, INPUT_HASH1, MERKLE_LEAVES1, signature, TXS1[:3], 0, GENESIS_BLOCK_HASH)
            ba.recv_sub_block_contender(TESTNET_DELEGATES[i]['vk'], sbc)

        self.assertFalse(ba.curr_block.is_consensus_reached())

        # self.assertEqual(len(ba.result_hashes[RESULT_HASH1]['_valid_signatures_']), DELEGATE_MAJORITY)
        # self.assertEqual(len(ba.result_hashes[RESULT_HASH1]['_transactions_']), 3)
        # self.assertEqual(len(ba.full_blocks), 0)


if __name__ == '__main__':
    unittest.main()



from .top import TopBlockManager
import os
import capnp
from cilantro_ee.core.messages.capnp_impl import capnp_struct as schemas
import bson
import hashlib

subblock_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/subblock.capnp')


def format_dictionary(d: dict) -> dict:
    for k, v in d.items():
        assert type(k) == str, 'Non-string key types not allowed.'
        if type(v) == list:
            for i in range(len(v)):
                if isinstance(v[i], dict):
                    v[i] = format_dictionary(v[i])
        elif isinstance(v, dict):
            d[k] = format_dictionary(v)
    return {k: v for k, v in sorted(d.items())}


def block_from_subblocks(subblocks, previous_hash: bytes, block_num: int) -> dict:
    block_hasher = hashlib.sha3_256()
    block_hasher.update(previous_hash)

    deserialized_subblocks = []

    for subblock in subblocks:
        sb = subblock.to_dict()

        sb = format_dictionary(sb)
        deserialized_subblocks.append(sb)

        encoded_sb = bson.BSON.encode(sb)
        block_hasher.update(encoded_sb)

    block = {
        'blockHash': block_hasher.digest(),
        'blockNum': block_num,
        'prevBlockHash': previous_hash,
        'subBlocks': deserialized_subblocks
    }

    return block


def verify_block(subblocks, previous_hash: bytes, proposed_hash: bytes):
    # Verify signatures!
    block_hasher = hashlib.sha3_256()
    block_hasher.update(previous_hash)

    deserialized_subblocks = []

    for subblock in subblocks:
        sb = subblock.to_dict()

        sb = format_dictionary(sb)
        deserialized_subblocks.append(sb)

        encoded_sb = bson.BSON.encode(sb)
        block_hasher.update(encoded_sb)

    if block_hasher.digest() == proposed_hash:
        return True

    return False



# Convenience
I = importlib

# Main state datum
contract_to_policy = Hash()
policy_to_contract = Hash()

# Policy interface
policy_interface = [
    I.Func('vote', args=('vk', 'obj')),
    I.Func('current_value')
]


@export
def register_policy(policy, contract):
    if policy_to_contract[policy] is None and contract_to_policy[contract] is None:
        # Attempt to import the contract to make sure it is already submitted
        p = I.import_module(contract)

        # Assert ownership is election_house and interface is correct
        assert I.owner_of(p) == ctx.this, \
            'Election house must control the policy contract!'

        assert I.enforce_interface(p, policy_interface), \
            'Policy contract does not follow the correct interface'


        policy_to_contract[policy] = contract
        contract_to_policy[contract] = policy
    else:
        raise Exception('Policy already registered')


@export
def current_value_for_policy(policy: str):
    contract = policy_to_contract[policy]
    assert contract is not None, 'Invalid policy.'

    p = I.import_module(contract)

    return p.current_value()


@export
def vote(policy, value):
    # Verify policy has been registered
    contract_name = policy_to_contract[policy]
    assert contract_name is not None, 'Invalid policy.'

    p = I.import_module(contract_name)

    p.vote(vk=ctx.caller, obj=value)



# from cilantro_ee.nodes.masternode.nonce import NonceManager
# from unittest import TestCase
#
#
# class TestNonce(TestCase):
#
#     def test_create_nonce(self):
#         user_vk = 'ABCD' * 16
#         nonce = NonceManager.create_nonce(user_vk)
#
#         self.assertTrue(user_vk in nonce)
#         self.assertEqual(len(nonce), 64*2 + 1)
#         self.assertTrue(NonceManager.check_if_exists(nonce))
#
#     def test_create_delete(self):
#         user_vk = 'ABCD' * 16
#         nonce = NonceManager.create_nonce(user_vk)
#
#         self.assertTrue(NonceManager.check_if_exists(nonce))
#         NonceManager.delete_nonce(nonce)
#         self.assertFalse(NonceManager.check_if_exists(nonce))
#



from unittest import TestCase
from contracting.webserver import app, client
import json
import aiohttp
import cilantro_ee

class TestWebserver(TestCase):
    def tearDown(self):
        client.flush()

    def test_ping_api(self):
        _, response = app.test_client.get('/')
        self.assertEqual(response.status, 200)

    def test_get_all_contracts(self):
        _, response = app.test_client.get('/contracts')
        contracts = response.json.get('contracts')
        self.assertListEqual(['submission'], contracts)

    def test_get_submission_code(self):
        with open(cilantro_ee.__path__[0] + '/contracts/submission.s.py') as f:
            contract = f.read()

        _, response = app.test_client.get('/contracts/submission')

        code = response.json.get('code')

        self.assertEqual(contract, code)
        self.assertEqual(response.status, 200)

    def test_get_non_existent_contract(self):
        _, response = app.test_client.get('/contracts/hoooooooopla')
        self.assertEqual(response.status, 404)

        error_message = response.json.get('error')

        self.assertEqual(error_message, 'hoooooooopla does not exist')

    def test_get_non_existent_contract_methods(self):
        _, response = app.test_client.get('/contracts/huuuuuuuuupluh/methods')

        self.assertEqual(response.status, 404)

        error_message = response.json.get('error')

        self.assertEqual(error_message, 'huuuuuuuuupluh does not exist')

    def test_get_variable_from_non_existent_contract(self):
        _, response = app.test_client.get('/contracts/currency/seed_amount')

        self.assertEqual(response.status, 404)

    def test_get_hash_from_non_existent_contract(self):
        _, response = app.test_client.get(
            '/contracts/currency/balances?key=324ee2e3544a8853a3c5a0ef0946b929aa488cbe7e7ee31a0fef9585ce398502'
        )

        self.assertEqual(response.status, 404)
        self.assertEqual(response.json.get('value'), None)




from unittest import TestCase
from cilantro_ee.contracts import sync
from contracting.db.driver import ContractDriver
from contracting.client import ContractingClient


class TestContractSync(TestCase):
    def test_directory_to_filename_works(self):
        directory = '~/something/something/hello/this/is/a/path.txt'
        name = 'path'
        _name = sync.contract_name_from_file_path(directory)

        self.assertEqual(name, _name)

    def test_directory_to_filename_if_just_filename(self):
        directory = 'path.txt'
        name = 'path'
        _name = sync.contract_name_from_file_path(directory)

        self.assertEqual(name, _name)

    def test_directory_to_filename_if_many_extensions(self):
        directory = 'path.txt.a.s.g.we.2.d.g.a.s.c.g'
        name = 'path'
        _name = sync.contract_name_from_file_path(directory)

        self.assertEqual(name, _name)

    def test_sync_genesis_contracts_if_none_in_instance(self):
        driver = ContractDriver()
        driver.flush()

        sync.sync_genesis_contracts()

        submission = driver.get_contract('submission')
        currency = driver.get_contract('currency')

        self.assertIsNotNone(submission)
        self.assertIsNotNone(currency)

    def test_sync_genesis_contracts_if_one_deleted(self):
        driver = ContractDriver()
        driver.flush()

        sync.sync_genesis_contracts()

        driver.delete_contract('submission')

        sync.sync_genesis_contracts()

        submission = driver.get_contract('submission')
        currency = driver.get_contract('currency')

        self.assertIsNotNone(submission)
        self.assertIsNotNone(currency)

    def test_sync_genesis_contracts_if_none_deleted(self):
        driver = ContractDriver()
        driver.flush()

        sync.sync_genesis_contracts()
        sync.sync_genesis_contracts()

        submission = driver.get_contract('submission')
        currency = driver.get_contract('currency')

        self.assertIsNotNone(submission)
        self.assertIsNotNone(currency)

    def test_submit_contract_with_specific_construction_args(self):
        driver = ContractDriver()
        driver.flush()

        sync.submit_contract_with_construction_args('vkbook', args={
            'masternodes': ['stu', 'raghu'],
            'delegates': ['tejas', 'monica'],
            'num_boot_mns': 1,
            'num_boot_del': 1
        })

        client = ContractingClient()
        vkbook = client.get_contract('vkbook')

        self.assertEqual(vkbook.get_masternodes(), ['stu', 'raghu'])
        self.assertEqual(vkbook.get_delegates(), ['tejas', 'monica'])



"""
General catchall for functions that don't make sense as methods.
"""
import hashlib
import asyncio


async def gather_dict(d):
    cors = list(d.values())
    results = await asyncio.gather(*cors)
    return dict(zip(d.keys(), results))


def digest(s):
    if not isinstance(s, bytes):
        s = str(s).encode('utf8')
    return hashlib.sha1(s).digest()


class OrderedSet(list):
    """
    Acts like a list in all ways, except in the behavior of the
    :meth:`push` method.
    """

    def push(self, thing):
        """
        1. If the item exists in the list, it's removed
        2. The item is pushed to the end of the list
        """
        if thing in self:
            self.remove(thing)
        self.append(thing)



from unittest import TestCase
from cilantro_ee.core.rewards import RewardManager
from tests import random_txs
from contracting.db.driver import ContractDriver
from contracting.client import ContractingClient
from cilantro_ee.contracts import genesis
import os
from cilantro_ee.services.storage.vkbook import VKBook, PhoneBook


class TestRewards(TestCase):
    def setUp(self):

        self.driver = ContractDriver()
        self.client = ContractingClient()

        genesis_path = os.path.dirname(genesis.__file__)

        with open(os.path.join(genesis_path, 'election_house.s.py')) as f:
            c = f.read()
            self.client.submit(c, name='election_house')

        with open(os.path.join(genesis_path, 'currency.s.py')) as f:
            c = f.read()
            self.client.submit(c, name='currency')

        with open(os.path.join(genesis_path, 'stamp_cost.s.py')) as f:
            c = f.read()
            self.client.submit(c, name='stamp_cost', owner='election_house', constructor_args={'initial_rate': 1_000_000})

        with open(os.path.join(genesis_path, 'rewards.s.py')) as f:
            c = f.read()
            self.client.submit(c, name='rewards', owner='election_house')

        self.r = RewardManager()

    def tearDown(self):
        self.driver.flush()

    def test_add_rewards(self):
        block = random_txs.random_block()

        total = 0

        for sb in block.subBlocks:
            for tx in sb.transactions:
                total += tx.stampsUsed

        self.assertEqual(self.r.stamps_in_block(block), total)

    def test_add_to_balance(self):
        currency_contract = self.client.get_contract('currency')
        current_balance = currency_contract.quick_read(variable='balances', key='test') or 0

        self.assertEqual(current_balance, 0)

        self.r.add_to_balance('test', 1234)

        current_balance = currency_contract.quick_read(variable='balances', key='test') or 0

        self.assertEqual(current_balance, 1234)

        self.r.add_to_balance('test', 1000)

        current_balance = currency_contract.quick_read(variable='balances', key='test') or 0

        self.assertEqual(current_balance, 2234)

    def test_stamps_per_tau_works(self):
        self.assertEqual(self.r.stamps_per_tau, 1_000_000)

        stamps = self.client.get_contract('stamp_cost')

        stamps.quick_write('S', 'rate', 555)

        self.assertEqual(self.r.stamps_per_tau, 555)

    def test_pending_rewards_get_sets(self):
        self.assertEqual(self.r.get_pending_rewards(), 0)

        self.r.set_pending_rewards(1000)

        self.assertEqual(self.r.get_pending_rewards(), 1000)

    def test_add_pending_rewards(self):
        block = random_txs.random_block()

        total = 0

        for tx in block.subBlocks[0].transactions:
            total += tx.stampsUsed

        expected = total / 1_000_000

        self.assertEqual(self.r.get_pending_rewards(), 0)

        self.r.add_pending_rewards(block.subBlocks[0])

        self.assertEqual(float(self.r.get_pending_rewards()), expected)

    def test_reward_ratio_works(self):
        self.assertEqual(self.r.reward_ratio, [0.5, 0.5, 0, 0])

    def test_issue_rewards_works(self):
        PhoneBook = VKBook(masternodes=['stu', 'raghu', 'steve'],
                           delegates=['tejas', 'alex'],
                           num_boot_mns=3,
                           num_boot_del=2,
                           stamps=True,
                           nonces=True)

        self.r.set_pending_rewards(1000)
        self.r.issue_rewards()

        currency_contract = self.client.get_contract('currency')

        self.r.add_to_balance('raghu', 1000)
        self.r.add_to_balance('steve', 10000)

        self.assertEqual(float(currency_contract.quick_read(variable='balances', key='stu')), 166.66666666666666)
        self.assertEqual(float(currency_contract.quick_read(variable='balances', key='raghu')), 1166.66666666666666)
        self.assertEqual(float(currency_contract.quick_read(variable='balances', key='steve')), 10166.66666666666666)

        self.assertEqual(currency_contract.quick_read(variable='balances', key='tejas'), 250)
        self.assertEqual(currency_contract.quick_read(variable='balances', key='alex'), 250)

        self.assertEqual(self.r.get_pending_rewards(), 0)



from cilantro_ee.core.logger.base import get_logger
import json
from cilantro_ee.constants import conf
from contracting.db.driver import RocksDriver
from contracting.db import encoder

import contextlib

from cilantro_ee.core.nonces import NonceManager


class MetaDataStorage(RocksDriver):
    def __init__(self,
                 block_hash_key='_current_block_hash',
                 epoch_hash_key='_current_epoch_hash',
                 block_num_key='_current_block_num',
                 nonce_key='__n',
                 pending_nonce_key='__pn'):

        self.block_hash_key = block_hash_key
        self.epoch_hash_key = epoch_hash_key
        self.block_num_key = block_num_key
        self.log = get_logger('StateDriver')
        self.interface = None

        self.nonce_key = nonce_key
        self.pending_nonce_key = pending_nonce_key

        self.nonce_manager = NonceManager()

        super().__init__()

    def get(self, key):
        value = super().get(key)
        return encoder.decode(value)

    def set(self, key, value):
        v = encoder.encode(value)
        super().set(key, v)

    def get_latest_block_hash(self):
        block_hash = super().get(self.block_hash_key)
        if block_hash is None:
            return b'\x00' * 32
        return block_hash

    def set_latest_block_hash(self, v: bytes):
        if type(v) == str:
            v = bytes.fromhex(v)
        assert len(v) == 32, 'Hash provided is not 32 bytes.'
        super().set(self.block_hash_key, v)

    latest_block_hash = property(get_latest_block_hash, set_latest_block_hash)

    def get_latest_epoch_hash(self):
        epoch_hash = super().get(self.epoch_hash_key)
        if epoch_hash is None:
            return b'\x00' * 32
        return epoch_hash

    def set_latest_epoch_hash(self, v: bytes):
        if type(v) == str:
            v = bytes.fromhex(v)
        assert len(v) == 32, 'Hash provided is not 32 bytes.'
        super().set(self.epoch_hash_key, v)

    latest_epoch_hash = property(get_latest_epoch_hash, set_latest_epoch_hash)

    def get_latest_block_num(self):
        num = self.get(self.block_num_key)
        if num is None:
            return 0

        return num

    def set_latest_block_num(self, v):
        v = int(v)
        assert v >= 0, 'Block number must be positive integer.'
        self.set(self.block_num_key, v)

    latest_block_num = property(get_latest_block_num, set_latest_block_num)

    def set_transaction_data(self, tx):
        if tx.state is not None and len(tx.state) > 0:
            with contextlib.suppress(json.JSONDecodeError):
                sets = json.loads(tx.state)

            if type(sets) != dict:
                return

            # For each KV in the JSON, set the key to the value
            for k, v in sets.items():
                self.set(k, v)

    def update_with_block(self, block):
        self.log.success('UPDATING STATE')

        # Capnp proto shim until we remove it completely from storage
        if type(block) != dict:
            block = block.to_dict()

        # self.log.info("block {}".format(block))

        assert self.latest_block_hash == block['prevBlockHash'], \
            "StateUpdate failed! Latest block hash {} does not match block data {}".format(self.latest_block_hash, block)

        # Map of tuple to nonce such that (processor, sender) => nonce
        nonces = {}

        for sb in block['subBlocks']:
            for tx in sb.transactions:
                self.nonce_manager.update_nonce_hash(nonce_hash=nonces, tx_payload=tx.transaction.payload)
                self.set_transaction_data(tx=tx)

        # Commit new nonces
        self.nonce_manager.commit_nonces(nonce_hash=nonces)
        self.nonce_manager.delete_pending_nonces()

        # Update our block hash and block num
        self.latest_block_hash = block['blockHash']
        self.latest_block_num = block['blockNum']

        # Update the epoch hash if it is time
        if self.latest_block_num % conf.EPOCH_INTERVAL == 0:
            self.latest_epoch_hash = self.latest_block_hash

            # Update rewards




def _lp_name(fn_name):
    """
    Helper function to get the internal lazy property name of a property
    :param fn: The name of the property function to be decorated
    :return: The property name prefixed by '_lazy_'
    """
    return '_lazy_' + fn_name


def set_lazy_property(obj, fn_name, value):
    """
    Convenience method to set the value of a lazy property before it is read. In the case that the value is known ahead
    of time, this should be implement to improve performance
    :param obj: The object the lazy property should be set on
    :param fn_name: The name of the property function to be set as a string
    :param value: The value to set the property to
    """
    attr_name = _lp_name(fn_name)
    setattr(obj, attr_name, value)


def lazy_property(fn):
    """
    Decorator that makes a property lazy-evaluated.
    :param fn: The property function to be decorated
    """
    attr_name = _lp_name(fn.__name__)

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property


def lazy_func(fn):
    """
    Slap this decorator on a func and memoize it (assumes func has no input or is called with same input every time)
    """
    func_name = _lp_name(fn.__name__)

    def _cache_func(self, *args, **kwargs):
        if not hasattr(self, func_name):
            setattr(self, func_name, fn(self, *args, **kwargs))
        return getattr(self, func_name)

    return _cache_func


# def set_lazy_func(obj, fn, value):
#     func_name = _lp_name(fn.__name__)
#
#     def _cache_func(*args, **kwargs):
#         return value
#
#     setattr(obj, func_name, _cache_func)



from contracting.db.driver import ContractDriver

BLOCK_HASH_KEY = '__last_block_hash'
BLOCK_NUMBER_KEY = '__last_block_number'


class TopBlockManager:
    def __init__(self, driver=ContractDriver()):
        self.driver = driver

    def get_latest_block_hash(self):
        return self.driver.get(BLOCK_HASH_KEY) or b'\x00' * 32

    def set_latest_block_hash(self, value):
        self.driver.set(BLOCK_HASH_KEY, value)

    def get_latest_block_number(self):
        return self.driver.get(BLOCK_NUMBER_KEY) or 0

    def set_latest_block_number(self, value):
        self.driver.set(BLOCK_NUMBER_KEY, value)



import zmq.asyncio
from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.core.crypto.wallet import Wallet, _verify
from cilantro_ee.core.sockets import services
from cilantro_ee.constants.ports import DISCOVERY_PORT
import asyncio
log = get_logger('DiscoveryService')

'''
DiscoverServer
Returns a message of the signed pepper and VK
'''


class DiscoveryServer(services.RequestReplyService):
    def __init__(self, socket_id: services.SocketStruct, wallet: Wallet, pepper: bytes, ctx=zmq.asyncio.Context(),
                 linger=2000, poll_timeout=500):

        super().__init__(socket_id=socket_id, wallet=wallet, ctx=ctx, linger=linger, poll_timeout=poll_timeout)

        self.pepper = pepper
        self.response = self.wallet.verifying_key() + self.wallet.sign(self.pepper)

    def handle_msg(self, msg):
        return self.response


def verify_vk_pepper(msg: bytes, pepper: bytes):
    if msg is None:
        return False

    assert len(msg) > 32, 'Message must be longer than 32 bytes.'
    vk, signed_pepper = unpack_pepper_msg(msg)
    return _verify(vk, pepper, signed_pepper)


def unpack_pepper_msg(msg: bytes):
    return msg[:32], msg[32:]


async def ping(socket_id: services.SocketStruct, pepper: bytes, ctx: zmq.Context, timeout):
    log.info('Timeout: {}'.format(timeout))
    response = await services.get(socket_id=socket_id, msg=b'', ctx=ctx, timeout=timeout)

    log.info('Got response: {}'.format(response))

    if verify_vk_pepper(response, pepper):
        log.info('Verifying key successfully extracted and message matches network pepper.')
        vk, _ = unpack_pepper_msg(response)
        return str(socket_id), vk

    return str(socket_id), None


async def discover_nodes(ip_list, pepper: bytes, ctx: zmq.Context, timeout=3000, retries=10):
    nodes_found = {}
    one_found = False
    retries_left = retries

    log.info([str(ip) for ip in ip_list])

    while not one_found and retries_left > 0:
        tasks = [ping(socket_id=ip, pepper=pepper, ctx=ctx, timeout=timeout) for ip in ip_list]

        tasks = asyncio.gather(*tasks)
        loop = asyncio.get_event_loop()

        log.info('Sending pings to {} nodes.'.format(len(ip_list)))

        if loop.is_running():
            results = await asyncio.ensure_future(tasks)
        else:
            results = loop.run_until_complete(tasks)

        for res in results:
            ip, vk = res
            if vk is not None:
                nodes_found[str(ip)] = vk.hex()
                one_found = True

        if not one_found:
            retries_left -= 1
            log.info('No one discovered... {} retried left.'.format(retries_left))

    # Returns mapping of IP -> VK. VKs that return None are not stored in the dictionary.
    return nodes_found




import os
import configparser
import cilantro_ee

settings = configparser.ConfigParser()
settings._interpolation = configparser.ExtendedInterpolation()
this_dir = os.path.dirname(__file__)
db_conf_path = os.path.join(cilantro_ee.__path__[0], 'services/storage/mn_db_conf.ini')

settings.read(db_conf_path)

# TODO deprecate ad-hoc stuff for VMNet. Use unified approach of loading settings from config file
# # If this is being run on VMNet, HOST_IP will be set. In this case we want to use preset DB_SETTINGS
# if os.getenv('HOST_IP'):
#     DB_SETTINGS = {'username': 'root',
#                    'password': '',
#                    'db': 'cilantro_ee_dev',
#                    'host': '127.0.0.1'
#                    }
# # Otherwise, this code is being run locally for unit test or on CI. In this case, we use our DB config file
# else:
DB_SETTINGS = {'username': settings.get('MN_DB', 'username'),
               'password': settings.get('MN_DB', 'password'),
               'db': settings.get('MN_DB', 'database'),
               'host': settings.get('MN_DB', 'hostname')
               }
